# vgp_startup_scripts_ext samba gpo policy
# Copyright (C) David Mulder <dmulder@suse.com> 2021
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from samba.gp.gpclass import gp_xml_ext, check_safe_path, gp_file_applier
from tempfile import NamedTemporaryFile
from samba.common import get_bytes
from subprocess import Popen

intro = b'''
### autogenerated by samba
#
# This file is generated by the vgp_startup_scripts_ext Group Policy
# Client Side Extension. To modify the contents of this file,
# modify the appropriate Group Policy objects which apply
# to this machine. DO NOT MODIFY THIS FILE DIRECTLY.
#

'''

class vgp_startup_scripts_ext(gp_xml_ext, gp_file_applier):
    def __str__(self):
        return 'VGP/Unix Settings/Startup Scripts'

    def process_group_policy(self, deleted_gpo_list, changed_gpo_list,
                             cdir='/etc/cron.d'):
        for guid, settings in deleted_gpo_list:
            if str(self) in settings:
                for attribute, script in settings[str(self)].items():
                    self.unapply(guid, attribute, script)

        for gpo in changed_gpo_list:
            if gpo.file_sys_path:
                xml = 'MACHINE/VGP/VTLA/Unix/Scripts/Startup/manifest.xml'
                path = os.path.join(gpo.file_sys_path, xml)
                xml_conf = self.parse(path)
                if not xml_conf:
                    continue
                policy = xml_conf.find('policysetting')
                data = policy.find('data')
                attributes = []
                for listelement in data.findall('listelement'):
                    local_path = self.lp.cache_path('gpo_cache')
                    script = listelement.find('script').text
                    script_file = os.path.join(local_path,
                        os.path.dirname(check_safe_path(path)).upper(),
                                        script.upper())
                    parameters = listelement.find('parameters')
                    if parameters is not None:
                        parameters = parameters.text
                    else:
                        parameters = ''
                    value_hash = listelement.find('hash').text
                    attribute = self.generate_attribute(script_file,
                                                        parameters)
                    attributes.append(attribute)
                    run_as = listelement.find('run_as')
                    if run_as is not None:
                        run_as = run_as.text
                    else:
                        run_as = 'root'
                    run_once = listelement.find('run_once') is not None
                    if run_once:
                        def applier_func(script_file, parameters):
                            Popen(['/bin/sh %s %s' % (script_file, parameters)],
                                shell=True).wait()
                            # Run once scripts don't create a file to unapply,
                            # so their is nothing to return.
                            return []
                        self.apply(gpo.name, attribute, value_hash, applier_func,
                                   script_file, parameters)
                    else:
                        def applier_func(run_as, script_file, parameters):
                            entry = '@reboot %s %s %s' % (run_as, script_file,
                                                          parameters)
                            with NamedTemporaryFile(prefix='gp_', dir=cdir,
                                                    delete=False) as f:
                                f.write(intro)
                                f.write(get_bytes(entry))
                                os.chmod(f.name, 0o700)
                                return [f.name]
                        self.apply(gpo.name, attribute, value_hash, applier_func,
                                   run_as, script_file, parameters)

                    self.clean(gpo.name, keep=attributes)

    def rsop(self, gpo):
        output = {}
        xml = 'MACHINE/VGP/VTLA/Unix/Scripts/Startup/manifest.xml'
        if gpo.file_sys_path:
            path = os.path.join(gpo.file_sys_path, xml)
            xml_conf = self.parse(path)
            if not xml_conf:
                return output
            policy = xml_conf.find('policysetting')
            data = policy.find('data')
            for listelement in data.findall('listelement'):
                local_path = self.lp.cache_path('gpo_cache')
                script = listelement.find('script').text
                script_file = os.path.join(local_path,
                    os.path.dirname(check_safe_path(path)).upper(),
                                    script.upper())
                parameters = listelement.find('parameters')
                if parameters is not None:
                    parameters = parameters.text
                else:
                    parameters = ''
                run_as = listelement.find('run_as')
                if run_as is not None:
                    run_as = run_as.text
                else:
                    run_as = 'root'
                run_once = listelement.find('run_once') is not None
                if run_once:
                    entry = 'Run once as: %s `%s %s`' % (run_as, script_file,
                                                         parameters)
                else:
                    entry = '@reboot %s %s %s' % (run_as, script_file,
                                                  parameters)
                if str(self) not in output.keys():
                    output[str(self)] = []
                output[str(self)].append(entry)
        return output



# Unix SMB/CIFS implementation.
# Copyright (C) Jelmer Vernooij <jelmer@samba.org> 2010
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for the libpolicy Python bindings.

"""

from samba.tests import TestCase
from samba import policy


class PolicyTests(TestCase):

    def test_get_gpo_flags(self):
        self.assertEqual(["GPO_FLAG_USER_DISABLE"],
                          policy.get_gpo_flags(policy.GPO_FLAG_USER_DISABLE))

    def test_get_gplink_options(self):
        self.assertEqual(["GPLINK_OPT_DISABLE"],
                          policy.get_gplink_options(policy.GPLINK_OPT_DISABLE))



# Unix SMB/CIFS implementation. Tests for dsdb
# Copyright (C) Matthieu Patou <mat@matws.net> 2010
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""These tests want to be run on a freshly provisioned domain that has
not been greatly modified by other tests (which at the time of writing
probably means 'chgdcpass').

Tests here should only read the database.

This is to avoid flapping tests.
"""

from samba.credentials import Credentials
from samba.samdb import SamDB
from samba.auth import system_session
from samba.tests import TestCase
from samba.gkdi import (
    KEY_CYCLE_DURATION,
    MAX_CLOCK_SKEW
)
from samba.nt_time import nt_now
import ldb
import samba


class DsdbQuietProvisionTests(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.lp = samba.tests.env_loadparm()
        cls.creds = Credentials()
        cls.creds.guess(cls.lp)
        cls.session = system_session()
        cls.samdb = SamDB(session_info=cls.session,
                          credentials=cls.creds,
                          lp=cls.lp)

    def test_dsdb_dn_gkdi_gmsa_root_keys_exist(self):
        """In provision we set up a GKDI root key.

        There should always be at least one that is already valid
        """
        current_time = nt_now()
        # We need the GKDI key to be already available for use
        min_use_start_time = current_time \
            - KEY_CYCLE_DURATION - MAX_CLOCK_SKEW

        dn = self.samdb.get_config_basedn()
        dn.add_child("CN=Master Root Keys,CN=Group Key Distribution Service,CN=Services")
        res = self.samdb.search(dn,
                                scope=ldb.SCOPE_SUBTREE,
                                expression=f"(&(objectClass = msKds-ProvRootKey)(msKds-UseStartTime<={min_use_start_time}))")

        self.assertGreater(len(res), 0)

    def test_dsdb_smartcard_expire_set(self):
        """In provision we set msDS-ExpirePasswordsOnSmartCardOnlyAccounts: TRUE for a new 2016 provision
        """
        dn = self.samdb.get_default_basedn()
        res = self.samdb.search(dn,
                                scope=ldb.SCOPE_BASE,
                                expression="(msDS-ExpirePasswordsOnSmartCardOnlyAccounts=TRUE)")

        self.assertEqual(len(res), 1)



# Unit and integration tests for traffic_packet.py
#
# Copyright (C) Catalyst IT Ltd. 2017
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import shutil
import tempfile


from samba.auth import system_session

from samba.credentials import MUST_USE_KERBEROS, DONT_USE_KERBEROS
from samba.emulate import traffic_packets as p
from samba.emulate import traffic
from samba.emulate.traffic import Packet

from samba.samdb import SamDB
import samba.tests
from samba import sd_utils


class TrafficEmulatorPacketTests(samba.tests.TestCase):
    def setUp(self):
        super().setUp()
        self.server      = os.environ["SERVER"]
        self.domain      = os.environ["DOMAIN"]
        self.host        = os.environ["SERVER_IP"]
        self.lp          = self.get_loadparm()
        self.session     = system_session()
        self.credentials = self.get_credentials()

        self.ldb = SamDB(url="ldap://%s" % self.host,
                         session_info=self.session,
                         credentials=self.credentials,
                         lp=self.lp)
        self.domain_sid = self.ldb.get_domain_sid()

        traffic.clean_up_accounts(self.ldb, 1)
        self.tempdir = tempfile.mkdtemp(prefix="traffic_packet_test_")
        self.context = traffic.ReplayContext(server=self.server,
                                             lp=self.lp,
                                             creds=self.credentials,
                                             tempdir=self.tempdir,
                                             ou=traffic.ou_name(self.ldb, 1),
                                             domain_sid=self.domain_sid,
                                             total_conversations=3,
                                             instance_id=1)

        self.conversation = traffic.Conversation()
        self.conversation.conversation_id = 1
        self.machinename = "STGM-1-1"
        self.machinepass = samba.generate_random_password(32, 32)
        self.username    = "STGU-1-1"
        self.userpass    = samba.generate_random_password(32, 32)
        account = traffic.ConversationAccounts(
            self.machinename,
            self.machinepass,
            self.username,
            self.userpass)

        traffic.create_ou(self.ldb, 1)
        traffic.create_machine_account(self.ldb,
                                       1,
                                       self.machinename,
                                       self.machinepass)
        traffic.create_user_account(self.ldb,
                                    1,
                                    self.username,
                                    self.userpass)

        self.context.generate_process_local_config(account, self.conversation)

        # grant user write permission to do things like write account SPN
        sdutils = sd_utils.SDUtils(self.ldb)
        mod = "(A;;WP;;;PS)"
        sdutils.dacl_add_ace(self.context.user_dn, mod)

    def tearDown(self):
        super().tearDown()
        traffic.clean_up_accounts(self.ldb, 1)
        del self.ldb
        shutil.rmtree(self.tempdir)

    def test_packet_cldap_03(self):
        packet = Packet.from_line(
            "0.0\t11\t1\t2\t1\tcldap\t3\tsearchRequest\t")
        self.assertTrue(p.packet_cldap_3(packet,
                                         self.conversation,
                                         self. context))

    def test_packet_cldap_05(self):
        packet = Packet.from_line(
            "0.0\t11\t1\t1\t2\tcldap\t5\tsearchResDone\t")
        self.assertFalse(p.packet_cldap_5(packet,
                                          self.conversation,
                                          self. context))

    def test_packet_dcerpc_00(self):
        packet = Packet.from_line("0.0\t11\t1\t2\t1\tdcerpc\t0\tRequest\t")
        self.assertFalse(p.packet_dcerpc_0(packet,
                                           self.conversation,
                                           self. context))

    def test_packet_dcerpc_02(self):
        packet = Packet.from_line("0.0\t11\t1\t1\t2\tdcerpc\t2\tResponse\t")
        self.assertFalse(p.packet_dcerpc_2(packet,
                                           self.conversation,
                                           self. context))

    def test_packet_dcerpc_03(self):
        packet = Packet.from_line("0.0\t11\t1\t1\t2\tdcerpc\t3\t\t")
        self.assertFalse(p.packet_dcerpc_3(packet,
                                           self.conversation,
                                           self. context))

    def test_packet_dcerpc_11(self):
        packet = Packet.from_line("0.0\t11\t1\t2\t1\tdcerpc\t11\tBind\t")
        self.assertFalse(p.packet_dcerpc_11(packet,
                                            self.conversation,
                                            self. context))

    def test_packet_dcerpc_13(self):
        packet = Packet.from_line("0.0\t11\t1\t2\t1\tdcerpc\t13\t\t")
        self.assertFalse(p.packet_dcerpc_13(packet,
                                            self.conversation,
                                            self. context))

    def test_packet_dcerpc_14(self):
        packet = Packet.from_line(
            "0.0\t11\t1\t2\t1\tdcerpc\t14\tAlter_context\t")
        self.assertFalse(p.packet_dcerpc_14(packet,
                                            self.conversation,
                                            self. context))

    def test_packet_dcerpc_15(self):
        packet = Packet.from_line(
            "0.0\t11\t1\t1\t2\tdcerpc\t15\tAlter_context_resp\t")
        # Set user_creds MUST_USE_KERBEROS to suppress the warning message.
        self.context.user_creds.set_kerberos_state(MUST_USE_KERBEROS)
        self.assertFalse(p.packet_dcerpc_15(packet,
                                            self.conversation,
                                            self. context))

    def test_packet_dcerpc_16(self):
        packet = Packet.from_line(
            "0.0\t11\t1\t1\t2\tdcerpc\t16\tAUTH3\t")
        self.assertFalse(p.packet_dcerpc_16(packet,
                                            self.conversation,
                                            self. context))

    def test_packet_dns_01(self):
        packet = Packet.from_line(
            "0.0\t11\t1\t1\t2\tdns\t1\tresponse\t")
        self.assertFalse(p.packet_dns_1(packet,
                                        self.conversation,
                                        self. context))

    def test_packet_drsuapi_00(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t1\t2\tdrsuapi\t0\tDsBind\t")
        self.assertTrue(p.packet_drsuapi_0(packet,
                                           self.conversation,
                                           self. context))

    def test_packet_drsuapi_01(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t1\t2\tdrsuapi\t1\tDsUnBind\t")
        self.assertTrue(p.packet_drsuapi_1(packet,
                                           self.conversation,
                                           self. context))

    def test_packet_drsuapi_02(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t1\t2\tdrsuapi\t2\tDsReplicaSync\t")
        self.assertFalse(p.packet_drsuapi_2(packet,
                                            self.conversation,
                                            self. context))

    def test_packet_drsuapi_03(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t1\t2\tdrsuapi\t3\tDsGetNCChanges\t")
        self.assertFalse(p.packet_drsuapi_3(packet,
                                            self.conversation,
                                            self. context))

    def test_packet_drsuapi_04(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t1\t2\tdrsuapi\t4\tDsReplicaUpdateRefs\t")
        self.assertFalse(p.packet_drsuapi_4(packet,
                                            self.conversation,
                                            self. context))

    def test_packet_drsuapi_12(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t1\t2\tdrsuapi\t12\tDsCrackNames\t")
        self.assertTrue(p.packet_drsuapi_12(packet,
                                            self.conversation,
                                            self. context))

    def test_packet_drsuapi_13(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t1\t2\tdrsuapi\t13\tDsWriteAccountSpn\t")
        self.assertTrue(p.packet_drsuapi_13(packet,
                                            self.conversation,
                                            self. context))

    def test_packet_epm_03(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t1\t2\tepm\t3\tMap\t")
        self.assertFalse(p.packet_epm_3(packet,
                                        self.conversation,
                                        self. context))

    def test_packet_kerberos(self):
        """Kerberos packets are not generated, but are used as a hint to
        favour kerberos.
        """
        packet = Packet.from_line(
            "0.0\t11\t1\t1\t2\tkerberos\t\t\t")
        self.assertFalse(p.packet_kerberos_(packet,
                                            self.conversation,
                                            self. context))
        self.assertEqual(MUST_USE_KERBEROS,
                         self.context.user_creds.get_kerberos_state())
        self.assertEqual(MUST_USE_KERBEROS,
                         self.context.user_creds_bad.get_kerberos_state())
        self.assertEqual(MUST_USE_KERBEROS,
                         self.context.machine_creds.get_kerberos_state())
        self.assertEqual(MUST_USE_KERBEROS,
                         self.context.machine_creds_bad.get_kerberos_state())
        self.assertEqual(MUST_USE_KERBEROS,
                         self.context.creds.get_kerberos_state())

        # Need to restore kerberos creds on the admin creds otherwise
        # subsequent tests fail
        self.credentials.set_kerberos_state(DONT_USE_KERBEROS)

    def test_packet_ldap(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t1\t2\tldap\t\t*** Unknown ***\t")
        self.assertFalse(p.packet_ldap_(packet,
                                        self.conversation,
                                        self. context))

    def test_packet_ldap_00_sasl(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tldap\t0\tbindRequest"
            "\t\t\t\t\t3\tsasl\t1.3.6.1.5.5.2")
        self.assertTrue(p.packet_ldap_0(packet,
                                        self.conversation,
                                        self. context))

    def test_packet_ldap_00_simple(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tldap\t0\tbindRequest"
            "\t\t\t\t\t0\tsimple\t")
        self.assertTrue(p.packet_ldap_0(packet,
                                        self.conversation,
                                        self. context))

    def test_packet_ldap_01(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t1\t2\tldap\t1\tbindResponse\t")
        self.assertFalse(p.packet_ldap_1(packet,
                                         self.conversation,
                                         self. context))

    def test_packet_ldap_02(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tldap\t2\tunbindRequest\t")
        self.assertFalse(p.packet_ldap_2(packet,
                                         self.conversation,
                                         self. context))

    def test_packet_ldap_03(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tldap\t3\tsearchRequest"
            "\t2\tDC,DC\t\tcn\t\t\t")
        self.assertTrue(p.packet_ldap_3(packet,
                                        self.conversation,
                                        self. context))

    def test_packet_ldap_04(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t1\t2\tldap\t4\tsearchResEntry\t")
        self.assertFalse(p.packet_ldap_4(packet,
                                         self.conversation,
                                         self. context))

    def test_packet_ldap_05(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t1\t2\tldap\t5\tsearchResDone\t")
        self.assertFalse(p.packet_ldap_5(packet,
                                         self.conversation,
                                         self. context))

    def test_packet_ldap_06(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tldap\t6\tmodifyRequest\t"
            "\t\t\t\t0\tadd")
        self.assertFalse(p.packet_ldap_6(packet,
                                         self.conversation,
                                         self. context))

    def test_packet_ldap_07(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t1\t2\tldap\t7\tmodifyResponse\t")
        self.assertFalse(p.packet_ldap_7(packet,
                                         self.conversation,
                                         self. context))

    def test_packet_ldap_08(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tldap\t8\taddRequest\t")
        self.assertFalse(p.packet_ldap_8(packet,
                                         self.conversation,
                                         self. context))

    def test_packet_ldap_09(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t1\t2\tldap\t9\taddResponse\t")
        self.assertFalse(p.packet_ldap_9(packet,
                                         self.conversation,
                                         self. context))

    def test_packet_ldap_16(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tldap\t16\tabandonRequest\t")
        self.assertFalse(p.packet_ldap_16(packet,
                                          self.conversation,
                                          self. context))

    def test_packet_lsarpc_00(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tlsarpc\t0\tlsa_Close\t")
        self.assertFalse(p.packet_lsarpc_1(packet,
                                           self.conversation,
                                           self. context))

    def test_packet_lsarpc_01(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tlsarpc\t1\tlsa_Delete\t")
        self.assertFalse(p.packet_lsarpc_1(packet,
                                           self.conversation,
                                           self. context))

    def test_packet_lsarpc_02(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tlsarpc\t2\tlsa_EnumeratePrivileges\t")
        self.assertFalse(p.packet_lsarpc_2(packet,
                                           self.conversation,
                                           self. context))

    def test_packet_lsarpc_03(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tlsarpc\t3\tlsa_QuerySecurityObject\t")
        self.assertFalse(p.packet_lsarpc_3(packet,
                                           self.conversation,
                                           self. context))

    def test_packet_lsarpc_04(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tlsarpc\t4\tlsa_SetSecurityObject\t")
        self.assertFalse(p.packet_lsarpc_4(packet,
                                           self.conversation,
                                           self. context))

    def test_packet_lsarpc_05(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tlsarpc\t5\tlsa_ChangePassword\t")
        self.assertFalse(p.packet_lsarpc_5(packet,
                                           self.conversation,
                                           self. context))

    def test_packet_lsarpc_06(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tlsarpc\t6\tlsa_OpenPolicy\t")
        self.assertFalse(p.packet_lsarpc_6(packet,
                                           self.conversation,
                                           self. context))

    def test_packet_lsarpc_14(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tlsarpc\t14\tlsa_LookupNames\t")
        self.assertTrue(p.packet_lsarpc_14(packet,
                                           self.conversation,
                                           self. context))

    def test_packet_lsarpc_15(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tlsarpc\t15\tlsa_LookupSids\t")
        self.assertTrue(p.packet_lsarpc_15(packet,
                                           self.conversation,
                                           self. context))

    def test_packet_lsarpc_39(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tlsarpc\t39\tlsa_QueryTrustedDomainInfoBySid\t")
        self.assertTrue(p.packet_lsarpc_39(packet,
                                           self.conversation,
                                           self. context))

    def test_packet_lsarpc_40(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tlsarpc\t40\tlsa_SetTrustedDomainInfo\t")
        self.assertFalse(p.packet_lsarpc_40(packet,
                                            self.conversation,
                                            self. context))

    def test_packet_lsarpc_43(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tlsarpc\t43\tlsa_StorePrivateData\t")
        self.assertFalse(p.packet_lsarpc_43(packet,
                                            self.conversation,
                                            self. context))

    def test_packet_lsarpc_44(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tlsarpc\t44\tlsa_RetrievePrivateData\t")
        self.assertFalse(p.packet_lsarpc_44(packet,
                                            self.conversation,
                                            self. context))

    def test_packet_lsarpc_68(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tlsarpc\t68\tlsa_LookupNames3\t")
        self.assertFalse(p.packet_lsarpc_68(packet,
                                            self.conversation,
                                            self. context))

    def test_packet_lsarpc_76(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tlsarpc\t76\tlsa_LookupSids3\t")
        self.assertTrue(p.packet_lsarpc_76(packet,
                                           self.conversation,
                                           self. context))

    def test_packet_lsarpc_77(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tlsarpc\t77\tlsa_LookupNames4\t")
        self.assertTrue(p.packet_lsarpc_77(packet,
                                           self.conversation,
                                           self. context))

    def test_packet_nbns_00(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tnbns\t0\tquery\t")
        self.assertTrue(p.packet_nbns_0(packet,
                                        self.conversation,
                                        self. context))

    def test_packet_nbns_01(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t1\t2\tnbns\t1\tresponse\t")
        self.assertTrue(p.packet_nbns_0(packet,
                                        self.conversation,
                                        self. context))

    def test_packet_rpc_netlogon_00(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\trpc_netlogon\t0\tNetrLogonUasLogon\t")
        self.assertFalse(p.packet_rpc_netlogon_0(packet,
                                                 self.conversation,
                                                 self. context))

    def test_packet_rpc_netlogon_01(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\trpc_netlogon\t1\tNetrLogonUasLogoff\t")
        self.assertFalse(p.packet_rpc_netlogon_1(packet,
                                                 self.conversation,
                                                 self. context))

    def test_packet_rpc_netlogon_04(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\trpc_netlogon\t4\tNetrServerReqChallenge\t")
        self.assertFalse(p.packet_rpc_netlogon_4(packet,
                                                 self.conversation,
                                                 self. context))

    def test_packet_rpc_netlogon_14(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\trpc_netlogon\t14\tNetrLogonControl2\t")
        self.assertFalse(p.packet_rpc_netlogon_14(packet,
                                                  self.conversation,
                                                  self. context))

    def test_packet_rpc_netlogon_15(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\trpc_netlogon\t15\tNetrServerAuthenticate2\t")
        self.assertFalse(p.packet_rpc_netlogon_15(packet,
                                                  self.conversation,
                                                  self. context))

    def test_packet_rpc_netlogon_21(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\trpc_netlogon\t21\tNetrLogonDummyRoutine1\t")
        self.assertFalse(p.packet_rpc_netlogon_21(packet,
                                                  self.conversation,
                                                  self. context))

    def test_packet_rpc_netlogon_26(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\trpc_netlogon\t26\tNetrServerAuthenticate3\t")
        self.assertFalse(p.packet_rpc_netlogon_26(packet,
                                                  self.conversation,
                                                  self. context))

    def test_packet_rpc_netlogon_29(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\trpc_netlogon\t29\tNetrLogonGetDomainInfo\t")
        self.assertTrue(p.packet_rpc_netlogon_29(packet,
                                                 self.conversation,
                                                 self. context))

    def test_packet_rpc_netlogon_30(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\trpc_netlogon\t30\tNetrServerPasswordSet2\t")
        self.assertTrue(p.packet_rpc_netlogon_30(packet,
                                                 self.conversation,
                                                 self. context))

    def test_packet_rpc_netlogon_34(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\trpc_netlogon\t34\tDsrGetDcNameEx2\t")
        self.assertFalse(p.packet_rpc_netlogon_34(packet,
                                                  self.conversation,
                                                  self. context))

    def test_packet_rpc_netlogon_39(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\trpc_netlogon\t39\tNetrLogonSamLogonEx\t")
        self.assertTrue(p.packet_rpc_netlogon_39(packet,
                                                 self.conversation,
                                                 self. context))

    def test_packet_rpc_netlogon_40(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\trpc_netlogon\t40\tDsrEnumerateDomainTrusts\t")
        self.assertTrue(p.packet_rpc_netlogon_40(packet,
                                                 self.conversation,
                                                 self. context))

    def test_packet_rpc_netlogon_45(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\trpc_netlogon\t45\tNetrLogonSamLogonWithFlags\t")
        self.assertTrue(p.packet_rpc_netlogon_45(packet,
                                                 self.conversation,
                                                 self. context))

    def test_packet_samr_00(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t0\tConnect\t")
        self.assertTrue(p.packet_samr_0(packet,
                                        self.conversation,
                                        self. context))

    def test_packet_samr_01(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t1\tClose\t")
        self.assertTrue(p.packet_samr_1(packet,
                                        self.conversation,
                                        self. context))

    def test_packet_samr_03(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t3\tQuerySecurity\t")
        self.assertTrue(p.packet_samr_3(packet,
                                        self.conversation,
                                        self. context))

    def test_packet_samr_05(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t5\tLookupDomain\t")
        self.assertTrue(p.packet_samr_5(packet,
                                        self.conversation,
                                        self. context))

    def test_packet_samr_06(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t6\tEnumDomains\t")
        self.assertTrue(p.packet_samr_6(packet,
                                        self.conversation,
                                        self. context))

    def test_packet_samr_07(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t7\tOpenDomain\t")
        self.assertTrue(p.packet_samr_7(packet,
                                        self.conversation,
                                        self. context))

    def test_packet_samr_08(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t8\tQueryDomainInfo'\t")
        self.assertTrue(p.packet_samr_8(packet,
                                        self.conversation,
                                        self. context))

    def test_packet_samr_14(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t14\tCreateDomAlias\t")
        self.assertFalse(p.packet_samr_14(packet,
                                          self.conversation,
                                          self. context))

    def test_packet_samr_15(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t15\tEnumDomainAliases\t")
        self.assertTrue(p.packet_samr_15(packet,
                                         self.conversation,
                                         self. context))

    def test_packet_samr_16(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t16\tGetAliasMembership\t")
        self.assertTrue(p.packet_samr_16(packet,
                                         self.conversation,
                                         self. context))

    def test_packet_samr_17(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t17\tLookupNames\t")
        self.assertTrue(p.packet_samr_17(packet,
                                         self.conversation,
                                         self. context))

    def test_packet_samr_18(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t18\tLookupRids\t")
        self.assertTrue(p.packet_samr_18(packet,
                                         self.conversation,
                                         self. context))

    def test_packet_samr_19(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t19\tOpenGroup\t")
        self.assertTrue(p.packet_samr_19(packet,
                                         self.conversation,
                                         self. context))

    def test_packet_samr_25(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t25\tQueryGroupMember\t")
        self.assertTrue(p.packet_samr_25(packet,
                                         self.conversation,
                                         self. context))

    def test_packet_samr_34(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t34\tOpenUser\t")
        self.assertTrue(p.packet_samr_34(packet,
                                         self.conversation,
                                         self. context))

    def test_packet_samr_36(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t36\tQueryUserInfo\t")
        self.assertTrue(p.packet_samr_36(packet,
                                         self.conversation,
                                         self. context))

    def test_packet_samr_37(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t37\tSetUserInfo\t")
        self.assertFalse(p.packet_samr_37(packet,
                                          self.conversation,
                                          self. context))

    def test_packet_samr_39(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t39\tGetGroupsForUser\t")
        self.assertTrue(p.packet_samr_39(packet,
                                         self.conversation,
                                         self. context))

    def test_packet_samr_40(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t40\tQueryDisplayInfo\t")
        self.assertFalse(p.packet_samr_40(packet,
                                          self.conversation,
                                          self. context))

    def test_packet_samr_44(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t44\tGetUserPwInfo\t")
        self.assertFalse(p.packet_samr_44(packet,
                                          self.conversation,
                                          self. context))

    def test_packet_samr_57(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t57\tConnect2\t")
        self.assertTrue(p.packet_samr_57(packet,
                                         self.conversation,
                                         self. context))

    def test_packet_samr_64(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t64\tConnect5\t")
        self.assertTrue(p.packet_samr_64(packet,
                                         self.conversation,
                                         self. context))

    def test_packet_samr_68(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsamr\t68\t\t")
        self.assertFalse(p.packet_samr_68(packet,
                                          self.conversation,
                                          self. context))

    def test_packet_srvsvc_16(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsrvsvc\t16\tNetShareGetInfo\t")
        self.assertTrue(p.packet_srvsvc_16(packet,
                                           self.conversation,
                                           self. context))

    def test_packet_srvsvc_21(self):
        packet = Packet.from_line(
            "0.0\t06\t1\t2\t1\tsrvsvc\t21\tNetSrvGetInfo\t")
        self.assertTrue(p.packet_srvsvc_21(packet,
                                           self.conversation,
                                           self. context))



# Unix SMB/CIFS implementation.
# Copyright (C) Andrew Bartlett <abartlet@samba.org> 2017
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for the Auth and AuthZ logging of password changes.
"""

import samba.tests
from samba.samdb import SamDB
from samba.auth import system_session
import os
import samba.tests.auth_log_base
from samba.tests import delete_force
from samba.net import Net
import samba
from ldb import LdbError
from samba.tests.password_test import PasswordCommon
from samba.dcerpc.windows_event_ids import (
    EVT_ID_SUCCESSFUL_LOGON,
    EVT_ID_UNSUCCESSFUL_LOGON,
    EVT_LOGON_NETWORK
)

USER_NAME = "authlogtestuser"
USER_PASS = samba.generate_random_password(32, 32)


class AuthLogPassChangeTests(samba.tests.auth_log_base.AuthLogTestBase):

    def setUp(self):
        super().setUp()

        self.server_ip = os.environ["SERVER_IP"]

        host = "ldap://%s" % os.environ["SERVER"]
        self.ldb = SamDB(url=host,
                         session_info=system_session(),
                         credentials=self.get_credentials(),
                         lp=self.get_loadparm())

        # permit password changes during this test
        PasswordCommon.allow_password_changes(self, self.ldb)

        self.base_dn = self.ldb.domain_dn()

        # (Re)adds the test user USER_NAME with password USER_PASS
        delete_force(self.ldb, "cn=" + USER_NAME + ",cn=users," + self.base_dn)
        self.ldb.add({
            "dn": "cn=" + USER_NAME + ",cn=users," + self.base_dn,
            "objectclass": "user",
            "sAMAccountName": USER_NAME,
            "userPassword": USER_PASS
        })

        # discard any auth log messages for the password setup
        type(self).discardMessages()

    def _authDescription(self):
        return "samr_ChangePasswordUser4"

    def test_admin_change_password(self):
        def isLastExpectedMessage(msg):
            return ((msg["type"] == "Authentication") and
                    (msg["Authentication"]["status"] == "NT_STATUS_OK") and
                    (msg["Authentication"]["serviceDescription"] ==
                        "SAMR Password Change") and
                    (msg["Authentication"]["authDescription"] ==
                        self._authDescription()) and
                    (msg["Authentication"]["eventId"] ==
                        EVT_ID_SUCCESSFUL_LOGON) and
                    (msg["Authentication"]["logonType"] ==
                        EVT_LOGON_NETWORK))

        creds = self.insta_creds(template=self.get_credentials())

        lp = self.get_loadparm()
        net = Net(creds, lp, server=self.server_ip)
        password = "newPassword!!42"

        net.change_password(newpassword=password,
                            username=USER_NAME,
                            oldpassword=USER_PASS)
        self.assertTrue(self.waitForMessages(isLastExpectedMessage),
                        "Did not receive the expected message")

    def test_admin_change_password_new_password_fails_restriction(self):
        def isLastExpectedMessage(msg):
            return ((msg["type"] == "Authentication") and
                    (msg["Authentication"]["status"] ==
                        "NT_STATUS_PASSWORD_RESTRICTION") and
                    (msg["Authentication"]["serviceDescription"] ==
                        "SAMR Password Change") and
                    (msg["Authentication"]["authDescription"] ==
                        self._authDescription()) and
                    (msg["Authentication"]["eventId"] ==
                        EVT_ID_UNSUCCESSFUL_LOGON) and
                    (msg["Authentication"]["logonType"] ==
                        EVT_LOGON_NETWORK))

        creds = self.insta_creds(template=self.get_credentials())

        lp = self.get_loadparm()
        net = Net(creds, lp, server=self.server_ip)
        password = "newPassword"

        exception_thrown = False
        try:
            net.change_password(newpassword=password,
                                oldpassword=USER_PASS,
                                username=USER_NAME)
        except Exception:
            exception_thrown = True
        self.assertEqual(True, exception_thrown,
                          "Expected exception not thrown")
        self.assertTrue(self.waitForMessages(isLastExpectedMessage),
                        "Did not receive the expected message")

    def test_admin_change_password_unknown_user(self):
        def isLastExpectedMessage(msg):
            return ((msg["type"] == "Authentication") and
                    (msg["Authentication"]["status"] ==
                        "NT_STATUS_NO_SUCH_USER") and
                    (msg["Authentication"]["serviceDescription"] ==
                        "SAMR Password Change") and
                    (msg["Authentication"]["authDescription"] ==
                        self._authDescription()) and
                    (msg["Authentication"]["eventId"] ==
                        EVT_ID_UNSUCCESSFUL_LOGON) and
                    (msg["Authentication"]["logonType"] ==
                        EVT_LOGON_NETWORK))

        creds = self.insta_creds(template=self.get_credentials())

        lp = self.get_loadparm()
        net = Net(creds, lp, server=self.server_ip)
        password = "newPassword!!42"

        exception_thrown = False
        try:
            net.change_password(newpassword=password,
                                oldpassword=USER_PASS,
                                username="badUser")
        except Exception:
            exception_thrown = True
        self.assertEqual(True, exception_thrown,
                          "Expected exception not thrown")

        self.assertTrue(self.waitForMessages(isLastExpectedMessage),
                        "Did not receive the expected message")

    def test_admin_change_password_bad_original_password(self):
        def isLastExpectedMessage(msg):
            return ((msg["type"] == "Authentication") and
                    (msg["Authentication"]["status"] ==
                        "NT_STATUS_WRONG_PASSWORD") and
                    (msg["Authentication"]["serviceDescription"] ==
                        "SAMR Password Change") and
                    (msg["Authentication"]["authDescription"] ==
                        self._authDescription()) and
                    (msg["Authentication"]["eventId"] ==
                        EVT_ID_UNSUCCESSFUL_LOGON) and
                    (msg["Authentication"]["logonType"] ==
                        EVT_LOGON_NETWORK))

        creds = self.insta_creds(template=self.get_credentials())

        lp = self.get_loadparm()
        net = Net(creds, lp, server=self.server_ip)
        password = "newPassword!!42"

        exception_thrown = False
        try:
            net.change_password(newpassword=password,
                                oldpassword="badPassword",
                                username=USER_NAME)
        except Exception:
            exception_thrown = True
        self.assertEqual(True, exception_thrown,
                          "Expected exception not thrown")

        self.assertTrue(self.waitForMessages(isLastExpectedMessage),
                        "Did not receive the expected message")

    def test_ldap_change_password(self):
        def isLastExpectedMessage(msg):
            return ((msg["type"] == "Authentication") and
                    (msg["Authentication"]["status"] == "NT_STATUS_OK") and
                    (msg["Authentication"]["serviceDescription"] ==
                        "LDAP Password Change") and
                    (msg["Authentication"]["authDescription"] ==
                        "LDAP Modify") and
                    (msg["Authentication"]["eventId"] ==
                        EVT_ID_SUCCESSFUL_LOGON) and
                    (msg["Authentication"]["logonType"] ==
                        EVT_LOGON_NETWORK))

        new_password = samba.generate_random_password(32, 32)
        self.ldb.modify_ldif(
            "dn: cn=" + USER_NAME + ",cn=users," + self.base_dn + "\n" +
            "changetype: modify\n" +
            "delete: userPassword\n" +
            "userPassword: " + USER_PASS + "\n" +
            "add: userPassword\n" +
            "userPassword: " + new_password + "\n")

        self.assertTrue(self.waitForMessages(isLastExpectedMessage),
                        "Did not receive the expected message")

    #
    # Currently this does not get logged, so we expect to see no messages.
    #
    def test_ldap_change_password_bad_user(self):
        def isLastExpectedMessage(msg):
            msg_type = msg["type"]

            # Accept any message we receive, except for those produced while
            # the Administrator authenticates in setUp().
            return (msg_type != "Authentication" or (
                "Administrator" not in msg[msg_type]["clientAccount"])) and (
                    msg_type != "Authorization" or (
                        "Administrator" not in msg[msg_type]["account"]))

        new_password = samba.generate_random_password(32, 32)
        try:
            self.ldb.modify_ldif(
                "dn: cn=" + "badUser" + ",cn=users," + self.base_dn + "\n" +
                "changetype: modify\n" +
                "delete: userPassword\n" +
                "userPassword: " + USER_PASS + "\n" +
                "add: userPassword\n" +
                "userPassword: " + new_password + "\n")
            self.fail()
        except LdbError as e:
            (num, msg) = e.args
            pass

        self.assertFalse(self.waitForMessages(isLastExpectedMessage),
                         "Received unexpected messages")

    def test_ldap_change_password_bad_original_password(self):
        def isLastExpectedMessage(msg):
            return ((msg["type"] == "Authentication") and
                    (msg["Authentication"]["status"] ==
                        "NT_STATUS_WRONG_PASSWORD") and
                    (msg["Authentication"]["serviceDescription"] ==
                        "LDAP Password Change") and
                    (msg["Authentication"]["authDescription"] ==
                        "LDAP Modify") and
                    (msg["Authentication"]["eventId"] ==
                        EVT_ID_UNSUCCESSFUL_LOGON) and
                    (msg["Authentication"]["logonType"] ==
                        EVT_LOGON_NETWORK))

        new_password = samba.generate_random_password(32, 32)
        try:
            self.ldb.modify_ldif(
                "dn: cn=" + USER_NAME + ",cn=users," + self.base_dn + "\n" +
                "changetype: modify\n" +
                "delete: userPassword\n" +
                "userPassword: " + "badPassword" + "\n" +
                "add: userPassword\n" +
                "userPassword: " + new_password + "\n")
            self.fail()
        except LdbError as e1:
            (num, msg) = e1.args
            pass

        self.assertTrue(self.waitForMessages(isLastExpectedMessage),
                        "Did not receive the expected message")



# Unix SMB/CIFS implementation.
# Copyright (C) Jelmer Vernooij <jelmer@samba.org> 2007
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for the _glue Python bindings."""

from samba import _glue
from samba import param
import samba.tests


class GlueTests(samba.tests.TestCase):

    def test_generate_random_str(self):
        string = _glue.generate_random_str(10)
        self.assertEqual(type(string), str)
        self.assertEqual(len(string), 10)

    def test_generate_random_password(self):
        password = _glue.generate_random_password(5, 10)
        self.assertEqual(type(password), str)
        self.assertTrue(5 <= len(password) <= 10)

    def test_unix2nttime(self):
        self.assertEqual(_glue.unix2nttime(1), 116444736010000000)

    def test_nttime2unix(self):
        self.assertEqual(_glue.nttime2unix(116444736010000000), 1)

    def test_float2nttime(self):
        self.assertEqual(_glue.float2nttime(1.0), 116444736010000000)
        self.assertEqual(_glue.float2nttime(1611058908.0), 132555325080000000)
        # NTTIME has a resolution of 100ns
        self.assertEqual(_glue.float2nttime(1611058908.1234567), 132555325081234567)
        self.assertEqual(_glue.float2nttime(1611058908.123456789), 132555325081234567)

    def test_nttime2float(self):
        self.assertEqual(_glue.nttime2float(1), -11644473600.0)
        self.assertEqual(_glue.nttime2float(0x7fffffffffffffff), 910692730085.4775)
        self.assertEqual(_glue.nttime2float(0x8000000000000000), 910692730085.4775)
        self.assertEqual(_glue.nttime2float(0xf000000000000000), 910692730085.4775)
        self.assertEqual(_glue.nttime2float(116444736010000000), 1.0)
        self.assertEqual(_glue.nttime2float(132555325080000000), 1611058908.0)
        self.assertEqual(_glue.nttime2float(132555325081234567), 1611058908.1234567)
        # NTTIME_OMIT (0) and NTTIME_FREEZE (UINT64_MAX) map to SAMBA_UTIME_OMIT (1)
        self.assertEqual(_glue.nttime2float(0), 1.0)
        self.assertEqual(_glue.nttime2float(0xffffffffffffffff), 1.0)

    def test_nttime2string(self):
        string = _glue.nttime2string(116444736010000000)
        self.assertEqual(type(string), str)
        self.assertIn('1970', string)

    def test_debug_level(self):
        prev_level = _glue.get_debug_level()
        try:
            self.assertIsNone(_glue.set_debug_level(0))
            self.assertEqual(_glue.get_debug_level(), 0)
            self.assertIsNone(_glue.set_debug_level(5))
            self.assertEqual(_glue.get_debug_level(), 5)
        finally:
            _glue.set_debug_level(prev_level)

    def test_interface_ips(self):
        lp = param.LoadParm()
        ips = _glue.interface_ips(lp)
        self.assertEqual(type(ips), list)

    def test_strcasecmp(self):
        self.assertEqual(_glue.strcasecmp_m('aA', 'Aa'), 0)
        self.assertNotEqual(_glue.strcasecmp_m('ab', 'Aa'), 0)

    def test_strstr_m(self):
        string = 'testing_string_num__one'
        self.assertEqual(_glue.strstr_m(string, '_'), '_string_num__one')
        self.assertEqual(_glue.strstr_m(string, '__'), '__one')
        self.assertEqual(_glue.strstr_m(string, 'ring'), 'ring_num__one')

    def test_crypt(self):
        # We hopefully only use schemes 5 and 6 (sha256 and sha512),
        # which are OK and also quite widely supported according to
        # https://en.wikipedia.org/wiki/Crypt_(C)
        for phrase, setting, expected in [
                ("a", "$5$aaaaaa",
                 "$5$aaaaaa$F4lxguL7mZR7TGlvukPTJIxoRhVmHMZs8ZdH8oDP0.6"),
                # with scheme 5, 5000 rounds is default, so hash is the same as above
                ('a', '$5$rounds=5000$aaaaaa',
                 '$5$rounds=5000$aaaaaa$F4lxguL7mZR7TGlvukPTJIxoRhVmHMZs8ZdH8oDP0.6'),
                ('a',
                 '$5$rounds=4999$aaaaaa',
                 '$5$rounds=4999$aaaaaa$FiP70gtxOJUFLokUJvET06E7jbL6aNmF6Wtv2ddzjY8'),
                ('a', '$5$aaaaab',
                 '$5$aaaaab$e9qR2F833/JyuMu.nkQc9kn184vBWLo0ODqnCe./mj0'),

                ('', '$5$aaaaaa', '$5$aaaaaa$5B4WTdWp5n/v/aNUw2N8RsEitqvlZJEaAKhH/pOkGg4'),

                ("a", "$6$aaaaaa",
                 "$6$aaaaaa$KHs/Ez7X/I5/K.V8FR7kEsx9rOvjXnEDUmGC.dLBWP87XWy.oUEAM7QYcZQRVhiDwGepOF2pKrCVETYLyASh60"),

                ('', '$5$', '$5$$3c2QQ0KjIU1OLtB29cl8Fplc2WN7X89bnoEjaR7tWu.'),

                # scheme 1 (md5) should be supported if not used
                ('a', '$1$aaaaaa',
                 '$1$aaaaaa$MUMWPbGfzrHFCNm7ZHg31.'),

                ('', '$6$',
                 '$6$$/chiBau24cE26QQVW3IfIe68Xu5.JQ4E8Ie7lcRLwqxO5cxGuBhqF2HmTL.zWJ9zjChg3yJYFXeGBQ2y3Ba1d1'),
                (' ',
                 '$6$6',
                 '$6$6$asLnbxf0obyuv3ybNvDE9ZcdwGFkDhLe7uW.wzdOdKCm4/M3vGFKq4Ttk1tBQrOn4wALZ3tj1L8IarIu5i8hR/'),

                # original DES scheme, 12 bits of salt
                ("a", "lalala", "laKGbFzgh./R2"),
                ("a", "lalalaLALALAla", "laKGbFzgh./R2"),
                ("a", "arrgh", "ar7VUiUvDhX2c"),
                ("a", "arrggghhh", "ar7VUiUvDhX2c"),
                ]:
            hash = _glue.crypt(phrase, setting)
            self.assertEqual(hash, expected)

    def test_crypt_bad(self):
        # We can't be too strident in our assertions, because every
        # system allows a different set of algorithms, and some have
        # different ideas of how to parse.
        for phrase, setting, exception in [
                ("a", "$5", ValueError),
                ("a", "$0$", ValueError),
                ("a", None, TypeError),
                (None, "", TypeError),
                ('a', '$66$', ValueError),
                ('a', '$$', ValueError),
                ('a', '*0', ValueError),
                ('a', '*', ValueError),
                ('a', '++', ValueError),
                # this next one is too long, except on Rocky Linux 8.
                #('a' * 10000, '$5$5', ValueError),
                # this is invalid, except on Debian 11.
                # (' ', '$6$ ', ValueError),
                ]:
            with self.assertRaises(exception,
                                   msg=f"crypt({phrase!r}, {setting!r}) didn't fail"):
                _glue.crypt(phrase, setting)



# domain management - domain keytab
#
# Copyright Matthias Dieter Wallnoefer 2009
# Copyright Andrew Kroeger 2009
# Copyright Jelmer Vernooij 2007-2012
# Copyright Giampaolo Lauria 2011
# Copyright Matthieu Patou <mat@matws.net> 2011
# Copyright Andrew Bartlett 2008-2015
# Copyright Stefan Metzmacher 2012
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import samba.getopt as options
from samba import enable_net_export_keytab, NTSTATUSError
from samba.net import Net
from samba.netcmd import Command, CommandError, Option

try:
    enable_net_export_keytab()
except ImportError:
    cmd_domain_export_keytab = None
else:
    class cmd_domain_export_keytab(Command):
        """Dump Kerberos keys of the domain into a keytab."""

        synopsis = "%prog <keytab> [options]"

        takes_optiongroups = {
            "sambaopts": options.SambaOptions,
            "credopts": options.CredentialsOptions,
            "hostopts": options.HostOptions,
            "versionopts": options.VersionOptions,
        }

        takes_options = [
            Option("--principal", help="extract only this principal", type=str),
            Option("--keep-stale-entries", help="keep stale keys in keytab (useful for collecting keys for Wireshark)", action="store_true"),
            Option("--only-current-keys",
                   help="This avoids exporting old and older keys (useful for keytabs used by kinit)",
                   action="store_true"),
        ]

        takes_args = ["keytab"]

        def run(self,
                keytab,
                credopts=None,
                sambaopts=None,
                versionopts=None,
                hostopts=None,
                principal=None,
                keep_stale_entries=None,
                only_current_keys=None):
            lp = sambaopts.get_loadparm()
            net = Net(None, lp)
            samdb = self.ldb_connect(hostopts, sambaopts, credopts)
            try:
                net.export_keytab(samdb=samdb,
                                  keytab=keytab,
                                  principal=principal,
                                  keep_stale_entries=keep_stale_entries,
                                  only_current_keys=only_current_keys)
            except NTSTATUSError as error:
                raise CommandError(f"Failed to export domain keys into keytab {keytab}: {error.args[1]}")



# Unix SMB/CIFS implementation. Tests for dsdb
# Copyright (C) Matthieu Patou <mat@matws.net> 2010
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for samba.dsdb."""

from samba.credentials import Credentials
from samba.samdb import SamDB
from samba.auth import system_session
from samba.tests import TestCase
from samba.tests import delete_force
from samba.ndr import ndr_unpack, ndr_pack
from samba.dcerpc import drsblobs, security, misc
from samba.param import LoadParm
from samba import dsdb, functional_level
from samba import werror
import ldb
import samba
import uuid


class DsdbAccountTests(TestCase):

    def setUp(self):
        super().setUp()
        self.lp = samba.tests.env_loadparm()
        self.creds = Credentials()
        self.creds.guess(self.lp)
        self.session = system_session()
        self.samdb = SamDB(session_info=self.session,
                           credentials=self.creds,
                           lp=self.lp)

        # Create a test user
        user_name = "dsdb-user-" + str(uuid.uuid4().hex[0:6])
        user_pass = samba.generate_random_password(32, 32)
        user_description = "Test user for dsdb test"

        base_dn = self.samdb.domain_dn()

        self.account_dn = "CN=" + user_name + ",CN=Users," + base_dn
        self.samdb.newuser(username=user_name,
                           password=user_pass,
                           description=user_description)
        # Cleanup (teardown)
        self.addCleanup(delete_force, self.samdb, self.account_dn)

        # Get server reference DN
        res = self.samdb.search(base=ldb.Dn(self.samdb,
                                            self.samdb.get_serverName()),
                                scope=ldb.SCOPE_BASE,
                                attrs=["serverReference"])
        # Get server reference
        self.server_ref_dn = ldb.Dn(
            self.samdb, res[0]["serverReference"][0].decode("utf-8"))

        # Get RID Set DN
        res = self.samdb.search(base=self.server_ref_dn,
                                scope=ldb.SCOPE_BASE,
                                attrs=["rIDSetReferences"])
        rid_set_refs = res[0]
        self.assertIn("rIDSetReferences", rid_set_refs)
        rid_set_str = rid_set_refs["rIDSetReferences"][0].decode("utf-8")
        self.rid_set_dn = ldb.Dn(self.samdb, rid_set_str)

    def get_rid_set(self, rid_set_dn):
        res = self.samdb.search(base=rid_set_dn,
                                scope=ldb.SCOPE_BASE,
                                attrs=["rIDAllocationPool",
                                       "rIDPreviousAllocationPool",
                                       "rIDUsedPool",
                                       "rIDNextRID"])
        return res[0]

    def test_ridalloc_next_free_rid(self):
        # Test RID allocation. We assume that RID
        # pools allocated to us are contiguous.
        self.samdb.transaction_start()
        try:
            orig_rid_set = self.get_rid_set(self.rid_set_dn)
            self.assertIn("rIDAllocationPool", orig_rid_set)
            self.assertIn("rIDPreviousAllocationPool", orig_rid_set)
            self.assertIn("rIDUsedPool", orig_rid_set)
            self.assertIn("rIDNextRID", orig_rid_set)

            # Get rIDNextRID value from RID set.
            next_rid = int(orig_rid_set["rIDNextRID"][0])

            # Check the result of next_free_rid().
            next_free_rid = self.samdb.next_free_rid()
            self.assertEqual(next_rid + 1, next_free_rid)

            # Check calling it twice in succession gives the same result.
            next_free_rid2 = self.samdb.next_free_rid()
            self.assertEqual(next_free_rid, next_free_rid2)

            # Ensure that the RID set attributes have not changed.
            rid_set2 = self.get_rid_set(self.rid_set_dn)
            self.assertEqual(orig_rid_set, rid_set2)
        finally:
            self.samdb.transaction_cancel()

    def test_ridalloc_no_ridnextrid(self):
        self.samdb.transaction_start()
        try:
            # Delete the rIDNextRID attribute of the RID set,
            # and set up previous and next pools.
            prev_lo = 1000
            prev_hi = 1999
            next_lo = 3000
            next_hi = 3999
            msg = ldb.Message()
            msg.dn = self.rid_set_dn
            msg["rIDNextRID"] = ldb.MessageElement([],
                                                   ldb.FLAG_MOD_DELETE,
                                                   "rIDNextRID")
            msg["rIDPreviousAllocationPool"] = (
                ldb.MessageElement(str((prev_hi << 32) | prev_lo),
                                   ldb.FLAG_MOD_REPLACE,
                                   "rIDPreviousAllocationPool"))
            msg["rIDAllocationPool"] = (
                ldb.MessageElement(str((next_hi << 32) | next_lo),
                                   ldb.FLAG_MOD_REPLACE,
                                   "rIDAllocationPool"))
            self.samdb.modify(msg)

            # Ensure that next_free_rid() returns the start of the next pool.
            next_free_rid3 = self.samdb.next_free_rid()
            self.assertEqual(next_lo, next_free_rid3)

            # Check the result of allocate_rid() matches.
            rid = self.samdb.allocate_rid()
            self.assertEqual(next_free_rid3, rid)

            # Check that the result of next_free_rid() has now changed.
            next_free_rid4 = self.samdb.next_free_rid()
            self.assertEqual(rid + 1, next_free_rid4)

            # Check the range of available RIDs.
            free_lo, free_hi = self.samdb.free_rid_bounds()
            self.assertEqual(rid + 1, free_lo)
            self.assertEqual(next_hi, free_hi)
        finally:
            self.samdb.transaction_cancel()

    def test_ridalloc_no_free_rids(self):
        self.samdb.transaction_start()
        try:
            # Exhaust our current pool of RIDs.
            pool_lo = 2000
            pool_hi = 2999
            msg = ldb.Message()
            msg.dn = self.rid_set_dn
            msg["rIDPreviousAllocationPool"] = (
                ldb.MessageElement(str((pool_hi << 32) | pool_lo),
                                   ldb.FLAG_MOD_REPLACE,
                                   "rIDPreviousAllocationPool"))
            msg["rIDAllocationPool"] = (
                ldb.MessageElement(str((pool_hi << 32) | pool_lo),
                                   ldb.FLAG_MOD_REPLACE,
                                   "rIDAllocationPool"))
            msg["rIDNextRID"] = (
            ldb.MessageElement(str(pool_hi),
                                   ldb.FLAG_MOD_REPLACE,
                                   "rIDNextRID"))
            self.samdb.modify(msg)

            # Ensure that calculating the next free RID fails.
            with self.assertRaises(ldb.LdbError) as err:
                self.samdb.next_free_rid()

            self.assertEqual("RID pools out of RIDs", err.exception.args[1])

            # Ensure we can still allocate a new RID.
            self.samdb.allocate_rid()
        finally:
            self.samdb.transaction_cancel()

    def test_ridalloc_new_ridset(self):
        self.samdb.transaction_start()
        try:
            # Test what happens with RID Set values set to zero (similar to
            # when a RID Set is first created, except we also set
            # rIDAllocationPool to zero).
            msg = ldb.Message()
            msg.dn = self.rid_set_dn
            msg["rIDPreviousAllocationPool"] = (
                ldb.MessageElement("0",
                                   ldb.FLAG_MOD_REPLACE,
                                   "rIDPreviousAllocationPool"))
            msg["rIDAllocationPool"] = (
                ldb.MessageElement("0",
                                   ldb.FLAG_MOD_REPLACE,
                                   "rIDAllocationPool"))
            msg["rIDNextRID"] = (
                ldb.MessageElement("0",
                                   ldb.FLAG_MOD_REPLACE,
                                   "rIDNextRID"))
            self.samdb.modify(msg)

            # Ensure that calculating the next free RID fails.
            with self.assertRaises(ldb.LdbError) as err:
                self.samdb.next_free_rid()

            self.assertEqual("RID pools out of RIDs", err.exception.args[1])

            # Set values for the next pool.
            pool_lo = 2000
            pool_hi = 2999
            msg = ldb.Message()
            msg.dn = self.rid_set_dn
            msg["rIDAllocationPool"] = (
                ldb.MessageElement(str((pool_hi << 32) | pool_lo),
                                   ldb.FLAG_MOD_REPLACE,
                                   "rIDAllocationPool"))
            self.samdb.modify(msg)

            # Ensure the next free RID value is equal to the next pool's lower
            # bound.
            next_free_rid5 = self.samdb.next_free_rid()
            self.assertEqual(pool_lo, next_free_rid5)

            # Check the range of available RIDs.
            free_lo, free_hi = self.samdb.free_rid_bounds()
            self.assertEqual(pool_lo, free_lo)
            self.assertEqual(pool_hi, free_hi)
        finally:
            self.samdb.transaction_cancel()

    def test_ridalloc_move_to_new_pool(self):
        self.samdb.transaction_start()
        try:
            # Test moving to a new pool from the previous pool.
            pool_lo = 2000
            pool_hi = 2999
            new_pool_lo = 4500
            new_pool_hi = 4599
            msg = ldb.Message()
            msg.dn = self.rid_set_dn
            msg["rIDPreviousAllocationPool"] = (
                ldb.MessageElement(str((pool_hi << 32) | pool_lo),
                                   ldb.FLAG_MOD_REPLACE,
                                   "rIDPreviousAllocationPool"))
            msg["rIDAllocationPool"] = (
                ldb.MessageElement(str((new_pool_hi << 32) | new_pool_lo),
                                   ldb.FLAG_MOD_REPLACE,
                                   "rIDAllocationPool"))
            msg["rIDNextRID"] = (
                ldb.MessageElement(str(pool_hi - 1),
                                   ldb.FLAG_MOD_REPLACE,
                                   "rIDNextRID"))
            self.samdb.modify(msg)

            # We should have remained in the previous pool.
            next_free_rid6 = self.samdb.next_free_rid()
            self.assertEqual(pool_hi, next_free_rid6)

            # Check the range of available RIDs.
            free_lo, free_hi = self.samdb.free_rid_bounds()
            self.assertEqual(pool_hi, free_lo)
            self.assertEqual(pool_hi, free_hi)

            # Allocate a new RID.
            rid2 = self.samdb.allocate_rid()
            self.assertEqual(next_free_rid6, rid2)

            # We should now move to the next pool.
            next_free_rid7 = self.samdb.next_free_rid()
            self.assertEqual(new_pool_lo, next_free_rid7)

            # Check the new range of available RIDs.
            free_lo2, free_hi2 = self.samdb.free_rid_bounds()
            self.assertEqual(new_pool_lo, free_lo2)
            self.assertEqual(new_pool_hi, free_hi2)

            # Ensure that allocate_rid() matches.
            rid3 = self.samdb.allocate_rid()
            self.assertEqual(next_free_rid7, rid3)
        finally:
            self.samdb.transaction_cancel()

    def test_ridalloc_no_ridsetreferences(self):
        self.samdb.transaction_start()
        try:
            # Delete the rIDSetReferences attribute.
            msg = ldb.Message()
            msg.dn = self.server_ref_dn
            msg["rIDSetReferences"] = (
                ldb.MessageElement([],
                                   ldb.FLAG_MOD_DELETE,
                                   "rIDSetReferences"))
            self.samdb.modify(msg)

            # Ensure calculating the next free RID fails.
            with self.assertRaises(ldb.LdbError) as err:
                self.samdb.next_free_rid()

            enum, estr = err.exception.args
            self.assertEqual(ldb.ERR_NO_SUCH_ATTRIBUTE, enum)
            self.assertIn("No RID Set DN - "
                          "Cannot find attribute rIDSetReferences of %s "
                          "to calculate reference dn" % self.server_ref_dn,
                          estr)

            # Ensure allocating a new RID fails.
            with self.assertRaises(ldb.LdbError) as err:
                self.samdb.allocate_rid()

            enum, estr = err.exception.args
            self.assertEqual(ldb.ERR_ENTRY_ALREADY_EXISTS, enum)
            self.assertIn("No RID Set DN - "
                          "Failed to add RID Set %s - "
                          "Entry %s already exists" %
                          (self.rid_set_dn, self.rid_set_dn),
                          estr)
        finally:
            self.samdb.transaction_cancel()

    def test_ridalloc_no_rid_set(self):
        self.samdb.transaction_start()
        try:
            # Set the rIDSetReferences attribute to not point to a RID Set.
            fake_rid_set_str = self.account_dn
            msg = ldb.Message()
            msg.dn = self.server_ref_dn
            msg["rIDSetReferences"] = (
                ldb.MessageElement(fake_rid_set_str,
                                   ldb.FLAG_MOD_REPLACE,
                                   "rIDSetReferences"))
            self.samdb.modify(msg)

            # Ensure calculating the next free RID fails.
            with self.assertRaises(ldb.LdbError) as err:
                self.samdb.next_free_rid()

            enum, estr = err.exception.args
            self.assertEqual(ldb.ERR_OPERATIONS_ERROR, enum)
            self.assertIn("Bad RID Set " + fake_rid_set_str, estr)

            # Ensure allocating a new RID fails.
            with self.assertRaises(ldb.LdbError) as err:
                self.samdb.allocate_rid()

            enum, estr = err.exception.args
            self.assertEqual(ldb.ERR_OPERATIONS_ERROR, enum)
            self.assertIn("Bad RID Set " + fake_rid_set_str,  estr)
        finally:
            self.samdb.transaction_cancel()

    def test_error_replpropertymetadata(self):
        res = self.samdb.search(scope=ldb.SCOPE_SUBTREE,
                                base=self.account_dn,
                                attrs=["replPropertyMetaData"])
        repl = ndr_unpack(drsblobs.replPropertyMetaDataBlob,
                          res[0]["replPropertyMetaData"][0])
        ctr = repl.ctr
        for o in ctr.array:
            # Search for Description
            if o.attid == 13:
                old_version = o.version
                o.version = o.version + 1
        replBlob = ndr_pack(repl)
        msg = ldb.Message()
        msg.dn = res[0].dn
        msg["replPropertyMetaData"] = ldb.MessageElement(replBlob, ldb.FLAG_MOD_REPLACE, "replPropertyMetaData")
        self.assertRaises(ldb.LdbError, self.samdb.modify, msg, ["local_oid:1.3.6.1.4.1.7165.4.3.14:0"])

    def test_error_replpropertymetadata_nochange(self):
        res = self.samdb.search(scope=ldb.SCOPE_SUBTREE,
                                base=self.account_dn,
                                attrs=["replPropertyMetaData"])
        repl = ndr_unpack(drsblobs.replPropertyMetaDataBlob,
                          res[0]["replPropertyMetaData"][0])
        replBlob = ndr_pack(repl)
        msg = ldb.Message()
        msg.dn = res[0].dn
        msg["replPropertyMetaData"] = ldb.MessageElement(replBlob, ldb.FLAG_MOD_REPLACE, "replPropertyMetaData")
        self.assertRaises(ldb.LdbError, self.samdb.modify, msg, ["local_oid:1.3.6.1.4.1.7165.4.3.14:0"])

    def test_error_replpropertymetadata_allow_sort(self):
        res = self.samdb.search(scope=ldb.SCOPE_SUBTREE,
                                base=self.account_dn,
                                attrs=["replPropertyMetaData"])
        repl = ndr_unpack(drsblobs.replPropertyMetaDataBlob,
                          res[0]["replPropertyMetaData"][0])
        replBlob = ndr_pack(repl)
        msg = ldb.Message()
        msg.dn = res[0].dn
        msg["replPropertyMetaData"] = ldb.MessageElement(replBlob, ldb.FLAG_MOD_REPLACE, "replPropertyMetaData")
        self.samdb.modify(msg, ["local_oid:1.3.6.1.4.1.7165.4.3.14:0", "local_oid:1.3.6.1.4.1.7165.4.3.25:0"])

    def test_twoatt_replpropertymetadata(self):
        res = self.samdb.search(scope=ldb.SCOPE_SUBTREE,
                                base=self.account_dn,
                                attrs=["replPropertyMetaData", "uSNChanged"])
        repl = ndr_unpack(drsblobs.replPropertyMetaDataBlob,
                          res[0]["replPropertyMetaData"][0])
        ctr = repl.ctr
        for o in ctr.array:
            # Search for Description
            if o.attid == 13:
                old_version = o.version
                o.version = o.version + 1
                o.local_usn = int(str(res[0]["uSNChanged"])) + 1
        replBlob = ndr_pack(repl)
        msg = ldb.Message()
        msg.dn = res[0].dn
        msg["replPropertyMetaData"] = ldb.MessageElement(replBlob, ldb.FLAG_MOD_REPLACE, "replPropertyMetaData")
        msg["description"] = ldb.MessageElement("new val", ldb.FLAG_MOD_REPLACE, "description")
        self.assertRaises(ldb.LdbError, self.samdb.modify, msg, ["local_oid:1.3.6.1.4.1.7165.4.3.14:0"])

    def test_set_replpropertymetadata(self):
        res = self.samdb.search(scope=ldb.SCOPE_SUBTREE,
                                base=self.account_dn,
                                attrs=["replPropertyMetaData", "uSNChanged"])
        repl = ndr_unpack(drsblobs.replPropertyMetaDataBlob,
                          res[0]["replPropertyMetaData"][0])
        ctr = repl.ctr
        for o in ctr.array:
            # Search for Description
            if o.attid == 13:
                old_version = o.version
                o.version = o.version + 1
                o.local_usn = int(str(res[0]["uSNChanged"])) + 1
                o.originating_usn = int(str(res[0]["uSNChanged"])) + 1
        replBlob = ndr_pack(repl)
        msg = ldb.Message()
        msg.dn = res[0].dn
        msg["replPropertyMetaData"] = ldb.MessageElement(replBlob, ldb.FLAG_MOD_REPLACE, "replPropertyMetaData")
        self.samdb.modify(msg, ["local_oid:1.3.6.1.4.1.7165.4.3.14:0"])

    def test_get_attribute_replmetadata_version(self):
        res = self.samdb.search(scope=ldb.SCOPE_SUBTREE,
                                base=self.account_dn,
                                attrs=["dn"])
        self.assertEqual(len(res), 1)
        dn = str(res[0].dn)
        self.assertEqual(self.samdb.get_attribute_replmetadata_version(dn, "unicodePwd"), 2)

    def test_set_attribute_replmetadata_version(self):
        res = self.samdb.search(scope=ldb.SCOPE_SUBTREE,
                                base=self.account_dn,
                                attrs=["dn"])
        self.assertEqual(len(res), 1)
        dn = str(res[0].dn)
        version = self.samdb.get_attribute_replmetadata_version(dn, "description")
        self.samdb.set_attribute_replmetadata_version(dn, "description", version + 2)
        self.assertEqual(self.samdb.get_attribute_replmetadata_version(dn, "description"), version + 2)

    def test_no_error_on_invalid_control(self):
        try:
            res = self.samdb.search(scope=ldb.SCOPE_SUBTREE,
                                    base=self.account_dn,
                                    attrs=["replPropertyMetaData"],
                                    controls=["local_oid:%s:0"
                                              % dsdb.DSDB_CONTROL_INVALID_NOT_IMPLEMENTED])
        except ldb.LdbError as e:
            self.fail("Should have not raised an exception")

    def test_error_on_invalid_critical_control(self):
        try:
            res = self.samdb.search(scope=ldb.SCOPE_SUBTREE,
                                    base=self.account_dn,
                                    attrs=["replPropertyMetaData"],
                                    controls=["local_oid:%s:1"
                                              % dsdb.DSDB_CONTROL_INVALID_NOT_IMPLEMENTED])
        except ldb.LdbError as e:
            (errno, estr) = e.args
            if errno != ldb.ERR_UNSUPPORTED_CRITICAL_EXTENSION:
                self.fail("Got %s should have got ERR_UNSUPPORTED_CRITICAL_EXTENSION"
                          % e[1])

class DsdbTests(TestCase):
    def setUp(self):
        super().setUp()
        self.lp = samba.tests.env_loadparm()
        self.creds = Credentials()
        self.creds.guess(self.lp)
        self.session = system_session()
        self.samdb = SamDB(session_info=self.session,
                           credentials=self.creds,
                           lp=self.lp)

    # Allocate a unique RID for use in the objectSID tests.
    #
    def allocate_rid(self):
        self.samdb.transaction_start()
        try:
            rid = self.samdb.allocate_rid()
        except:
            self.samdb.transaction_cancel()
            raise
        self.samdb.transaction_commit()
        return str(rid)

    def test_get_oid_from_attrid(self):
        oid = self.samdb.get_oid_from_attid(591614)
        self.assertEqual(oid, "1.2.840.113556.1.4.1790")

    def test_ok_get_attribute_from_attid(self):
        self.assertEqual(self.samdb.get_attribute_from_attid(13), "description")

    def test_ko_get_attribute_from_attid(self):
        self.assertEqual(self.samdb.get_attribute_from_attid(11979), None)

    # Ensure that duplicate objectSID's are permitted for foreign security
    # principals.
    #
    def test_duplicate_objectSIDs_allowed_on_foreign_security_principals(self):

        #
        # We need to build a foreign security principal SID
        # i.e a  SID not in the current domain.
        #
        dom_sid = self.samdb.get_domain_sid()
        if str(dom_sid).endswith("0"):
            c = "9"
        else:
            c = "0"
        sid_str = str(dom_sid)[:-1] + c + "-1000"
        sid     = ndr_pack(security.dom_sid(sid_str))
        basedn  = self.samdb.get_default_basedn()
        dn      = "CN=%s,CN=ForeignSecurityPrincipals,%s" % (sid_str, basedn)

        #
        # First without control
        #

        try:
            self.samdb.add({
                "dn": dn,
                "objectClass": "foreignSecurityPrincipal"})
            self.fail("No exception should get ERR_OBJECT_CLASS_VIOLATION")
        except ldb.LdbError as e:
            (code, msg) = e.args
            self.assertEqual(code, ldb.ERR_OBJECT_CLASS_VIOLATION, str(e))
            werr = "%08X" % werror.WERR_DS_MISSING_REQUIRED_ATT
            self.assertTrue(werr in msg, msg)

        try:
            self.samdb.add({
                "dn": dn,
                "objectClass": "foreignSecurityPrincipal",
                "objectSid": sid})
            self.fail("No exception should get ERR_UNWILLING_TO_PERFORM")
        except ldb.LdbError as e:
            (code, msg) = e.args
            self.assertEqual(code, ldb.ERR_UNWILLING_TO_PERFORM, str(e))
            werr = "%08X" % werror.WERR_DS_ILLEGAL_MOD_OPERATION
            self.assertTrue(werr in msg, msg)

        #
        # We need to use the provision control
        # in order to add foreignSecurityPrincipal
        # objects
        #

        controls = ["provision:0"]
        self.samdb.add({
            "dn": dn,
            "objectClass": "foreignSecurityPrincipal"},
            controls=controls)

        self.samdb.delete(dn)

        try:
            self.samdb.add({
                "dn": dn,
                "objectClass": "foreignSecurityPrincipal"},
                controls=controls)
        except ldb.LdbError as e:
            (code, msg) = e.args
            self.fail("Got unexpected exception %d - %s "
                      % (code, msg))

        # cleanup
        self.samdb.delete(dn)

    def _test_foreignSecurityPrincipal(self, obj_class, fpo_attr):

        dom_sid = self.samdb.get_domain_sid()
        lsid_str = str(dom_sid) + "-4294967294"
        bsid_str = "S-1-5-32-4294967294"
        fsid_str = "S-1-5-4294967294"
        basedn   = self.samdb.get_default_basedn()
        cn       = "dsdb_test_fpo"
        dn_str   = "cn=%s,cn=Users,%s" % (cn, basedn)
        dn = ldb.Dn(self.samdb, dn_str)

        res = self.samdb.search(scope=ldb.SCOPE_SUBTREE,
                                base=basedn,
                                expression="(objectSid=%s)" % lsid_str,
                                attrs=[])
        self.assertEqual(len(res), 0)
        res = self.samdb.search(scope=ldb.SCOPE_SUBTREE,
                                base=basedn,
                                expression="(objectSid=%s)" % bsid_str,
                                attrs=[])
        self.assertEqual(len(res), 0)
        res = self.samdb.search(scope=ldb.SCOPE_SUBTREE,
                                base=basedn,
                                expression="(objectSid=%s)" % fsid_str,
                                attrs=[])
        self.assertEqual(len(res), 0)

        self.addCleanup(delete_force, self.samdb, dn_str)

        self.samdb.add({
            "dn": dn_str,
            "objectClass": obj_class})

        msg = ldb.Message()
        msg.dn = dn
        msg[fpo_attr] = ldb.MessageElement("<SID=%s>" % lsid_str,
                                           ldb.FLAG_MOD_ADD,
                                           fpo_attr)
        try:
            self.samdb.modify(msg)
            self.fail("No exception should get LDB_ERR_UNWILLING_TO_PERFORM")
        except ldb.LdbError as e:
            (code, msg) = e.args
            self.assertEqual(code, ldb.ERR_UNWILLING_TO_PERFORM, str(e))
            werr = "%08X" % werror.WERR_DS_INVALID_GROUP_TYPE
            self.assertTrue(werr in msg, msg)

        msg = ldb.Message()
        msg.dn = dn
        msg[fpo_attr] = ldb.MessageElement("<SID=%s>" % bsid_str,
                                           ldb.FLAG_MOD_ADD,
                                           fpo_attr)
        try:
            self.samdb.modify(msg)
            self.fail("No exception should get LDB_ERR_NO_SUCH_OBJECT")
        except ldb.LdbError as e:
            (code, msg) = e.args
            self.assertEqual(code, ldb.ERR_NO_SUCH_OBJECT, str(e))
            werr = "%08X" % werror.WERR_NO_SUCH_MEMBER
            self.assertTrue(werr in msg, msg)

        msg = ldb.Message()
        msg.dn = dn
        msg[fpo_attr] = ldb.MessageElement("<SID=%s>" % fsid_str,
                                           ldb.FLAG_MOD_ADD,
                                           fpo_attr)
        try:
            self.samdb.modify(msg)
        except ldb.LdbError as e:
            self.fail("Should have not raised an exception")

        res = self.samdb.search(scope=ldb.SCOPE_SUBTREE,
                                base=basedn,
                                expression="(objectSid=%s)" % fsid_str,
                                attrs=[])
        self.assertEqual(len(res), 1)
        self.samdb.delete(res[0].dn)
        self.samdb.delete(dn)
        res = self.samdb.search(scope=ldb.SCOPE_SUBTREE,
                                base=basedn,
                                expression="(objectSid=%s)" % fsid_str,
                                attrs=[])
        self.assertEqual(len(res), 0)

    def test_foreignSecurityPrincipal_member(self):
        return self._test_foreignSecurityPrincipal(
                "group", "member")

    def test_foreignSecurityPrincipal_MembersForAzRole(self):
        return self._test_foreignSecurityPrincipal(
                "msDS-AzRole", "msDS-MembersForAzRole")

    def test_foreignSecurityPrincipal_NeverRevealGroup(self):
        return self._test_foreignSecurityPrincipal(
                "computer", "msDS-NeverRevealGroup")

    def test_foreignSecurityPrincipal_RevealOnDemandGroup(self):
        return self._test_foreignSecurityPrincipal(
                "computer", "msDS-RevealOnDemandGroup")

    def _test_fail_foreignSecurityPrincipal(self, obj_class, fpo_attr,
                                            msg_exp, lerr_exp, werr_exp,
                                            allow_reference=True):

        dom_sid = self.samdb.get_domain_sid()
        lsid_str = str(dom_sid) + "-4294967294"
        bsid_str = "S-1-5-32-4294967294"
        fsid_str = "S-1-5-4294967294"
        basedn   = self.samdb.get_default_basedn()
        cn1       = "dsdb_test_fpo1"
        dn1_str   = "cn=%s,cn=Users,%s" % (cn1, basedn)
        dn1 = ldb.Dn(self.samdb, dn1_str)
        cn2       = "dsdb_test_fpo2"
        dn2_str   = "cn=%s,cn=Users,%s" % (cn2, basedn)
        dn2 = ldb.Dn(self.samdb, dn2_str)

        res = self.samdb.search(scope=ldb.SCOPE_SUBTREE,
                                base=basedn,
                                expression="(objectSid=%s)" % lsid_str,
                                attrs=[])
        self.assertEqual(len(res), 0)
        res = self.samdb.search(scope=ldb.SCOPE_SUBTREE,
                                base=basedn,
                                expression="(objectSid=%s)" % bsid_str,
                                attrs=[])
        self.assertEqual(len(res), 0)
        res = self.samdb.search(scope=ldb.SCOPE_SUBTREE,
                                base=basedn,
                                expression="(objectSid=%s)" % fsid_str,
                                attrs=[])
        self.assertEqual(len(res), 0)

        self.addCleanup(delete_force, self.samdb, dn1_str)
        self.addCleanup(delete_force, self.samdb, dn2_str)

        self.samdb.add({
            "dn": dn1_str,
            "objectClass": obj_class})

        self.samdb.add({
            "dn": dn2_str,
            "objectClass": obj_class})

        msg = ldb.Message()
        msg.dn = dn1
        msg[fpo_attr] = ldb.MessageElement("<SID=%s>" % lsid_str,
                                           ldb.FLAG_MOD_ADD,
                                           fpo_attr)
        try:
            self.samdb.modify(msg)
            self.fail("No exception should get %s" % msg_exp)
        except ldb.LdbError as e:
            (code, msg) = e.args
            self.assertEqual(code, lerr_exp, str(e))
            werr = "%08X" % werr_exp
            self.assertTrue(werr in msg, msg)

        msg = ldb.Message()
        msg.dn = dn1
        msg[fpo_attr] = ldb.MessageElement("<SID=%s>" % bsid_str,
                                           ldb.FLAG_MOD_ADD,
                                           fpo_attr)
        try:
            self.samdb.modify(msg)
            self.fail("No exception should get %s" % msg_exp)
        except ldb.LdbError as e:
            (code, msg) = e.args
            self.assertEqual(code, lerr_exp, str(e))
            werr = "%08X" % werr_exp
            self.assertTrue(werr in msg, msg)

        msg = ldb.Message()
        msg.dn = dn1
        msg[fpo_attr] = ldb.MessageElement("<SID=%s>" % fsid_str,
                                           ldb.FLAG_MOD_ADD,
                                           fpo_attr)
        try:
            self.samdb.modify(msg)
            self.fail("No exception should get %s" % msg)
        except ldb.LdbError as e:
            (code, msg) = e.args
            self.assertEqual(code, lerr_exp, str(e))
            werr = "%08X" % werr_exp
            self.assertTrue(werr in msg, msg)

        msg = ldb.Message()
        msg.dn = dn1
        msg[fpo_attr] = ldb.MessageElement("%s" % dn2,
                                           ldb.FLAG_MOD_ADD,
                                           fpo_attr)
        try:
            self.samdb.modify(msg)
            if not allow_reference:
                self.fail("No exception should get %s" % msg_exp)
        except ldb.LdbError as e:
            if allow_reference:
                self.fail("Should have not raised an exception: %s" % e)
            (code, msg) = e.args
            self.assertEqual(code, lerr_exp, str(e))
            werr = "%08X" % werr_exp
            self.assertTrue(werr in msg, msg)

        self.samdb.delete(dn2)
        self.samdb.delete(dn1)

    def test_foreignSecurityPrincipal_NonMembers(self):
        return self._test_fail_foreignSecurityPrincipal(
                "group", "msDS-NonMembers",
                "LDB_ERR_UNWILLING_TO_PERFORM/WERR_NOT_SUPPORTED",
                ldb.ERR_UNWILLING_TO_PERFORM, werror.WERR_NOT_SUPPORTED,
                allow_reference=False)

    def test_foreignSecurityPrincipal_HostServiceAccount(self):
        return self._test_fail_foreignSecurityPrincipal(
                "computer", "msDS-HostServiceAccount",
                "LDB_ERR_CONSTRAINT_VIOLATION/WERR_DS_NAME_REFERENCE_INVALID",
                ldb.ERR_CONSTRAINT_VIOLATION,
                werror.WERR_DS_NAME_REFERENCE_INVALID)

    def test_foreignSecurityPrincipal_manager(self):
        return self._test_fail_foreignSecurityPrincipal(
                "user", "manager",
                "LDB_ERR_CONSTRAINT_VIOLATION/WERR_DS_NAME_REFERENCE_INVALID",
                ldb.ERR_CONSTRAINT_VIOLATION,
                werror.WERR_DS_NAME_REFERENCE_INVALID)

    #
    # Duplicate objectSID's should not be permitted for sids in the local
    # domain. The test sequence is add an object, delete it, then attempt to
    # re-add it, this should fail with a constraint violation
    #
    def test_duplicate_objectSIDs_not_allowed_on_local_objects(self):

        dom_sid = self.samdb.get_domain_sid()
        rid     = self.allocate_rid()
        sid_str = str(dom_sid) + "-" + rid
        sid     = ndr_pack(security.dom_sid(sid_str))
        basedn  = self.samdb.get_default_basedn()
        cn       = "dsdb_test_01"
        dn      = "cn=%s,cn=Users,%s" % (cn, basedn)

        self.samdb.add({
            "dn": dn,
            "objectClass": "user",
            "objectSID": sid})
        self.samdb.delete(dn)

        try:
            self.samdb.add({
                "dn": dn,
                "objectClass": "user",
                "objectSID": sid})
            self.fail("No exception should get LDB_ERR_CONSTRAINT_VIOLATION")
        except ldb.LdbError as e:
            (code, msg) = e.args
            if code != ldb.ERR_CONSTRAINT_VIOLATION:
                self.fail("Got %d - %s should have got "
                          "LDB_ERR_CONSTRAINT_VIOLATION"
                          % (code, msg))

    def test_linked_vs_non_linked_reference(self):
        basedn   = self.samdb.get_default_basedn()
        kept_dn_str   = "cn=reference_kept,cn=Users,%s" % (basedn)
        removed_dn_str   = "cn=reference_removed,cn=Users,%s" % (basedn)
        dom_sid = self.samdb.get_domain_sid()
        none_sid_str = str(dom_sid) + "-4294967294"
        none_guid_str = "afafafaf-fafa-afaf-fafa-afafafafafaf"

        self.addCleanup(delete_force, self.samdb, kept_dn_str)
        self.addCleanup(delete_force, self.samdb, removed_dn_str)

        self.samdb.add({
            "dn": kept_dn_str,
            "objectClass": "user"})
        res = self.samdb.search(scope=ldb.SCOPE_SUBTREE,
                                base=kept_dn_str,
                                attrs=["objectGUID", "objectSID"])
        self.assertEqual(len(res), 1)
        kept_guid = ndr_unpack(misc.GUID, res[0]["objectGUID"][0])
        kept_sid = ndr_unpack(security.dom_sid, res[0]["objectSid"][0])
        kept_dn = res[0].dn

        self.samdb.add({
            "dn": removed_dn_str,
            "objectClass": "user"})
        res = self.samdb.search(scope=ldb.SCOPE_SUBTREE,
                                base=removed_dn_str,
                                attrs=["objectGUID", "objectSID"])
        self.assertEqual(len(res), 1)
        removed_guid = ndr_unpack(misc.GUID, res[0]["objectGUID"][0])
        removed_sid = ndr_unpack(security.dom_sid, res[0]["objectSid"][0])
        self.samdb.delete(removed_dn_str)

        #
        # First try the linked attribute 'manager'
        # by GUID and SID
        #

        msg = ldb.Message()
        msg.dn = kept_dn
        msg["manager"] = ldb.MessageElement("<SID=%s>" % removed_sid,
                                            ldb.FLAG_MOD_ADD,
                                            "manager")
        try:
            self.samdb.modify(msg)
            self.fail("No exception should get LDB_ERR_CONSTRAINT_VIOLATION")
        except ldb.LdbError as e:
            (code, msg) = e.args
            self.assertEqual(code, ldb.ERR_CONSTRAINT_VIOLATION, str(e))
            werr = "%08X" % werror.WERR_DS_NAME_REFERENCE_INVALID
            self.assertTrue(werr in msg, msg)

        msg = ldb.Message()
        msg.dn = kept_dn
        msg["manager"] = ldb.MessageElement("<GUID=%s>" % removed_guid,
                                            ldb.FLAG_MOD_ADD,
                                            "manager")
        try:
            self.samdb.modify(msg)
            self.fail("No exception should get LDB_ERR_CONSTRAINT_VIOLATION")
        except ldb.LdbError as e:
            (code, msg) = e.args
            self.assertEqual(code, ldb.ERR_CONSTRAINT_VIOLATION, str(e))
            werr = "%08X" % werror.WERR_DS_NAME_REFERENCE_INVALID
            self.assertTrue(werr in msg, msg)

        #
        # Try the non-linked attribute 'assistant'
        # by GUID and SID, which should work.
        #
        msg = ldb.Message()
        msg.dn = kept_dn
        msg["assistant"] = ldb.MessageElement("<SID=%s>" % removed_sid,
                                              ldb.FLAG_MOD_ADD,
                                              "assistant")
        self.samdb.modify(msg)
        msg = ldb.Message()
        msg.dn = kept_dn
        msg["assistant"] = ldb.MessageElement("<SID=%s>" % removed_sid,
                                              ldb.FLAG_MOD_DELETE,
                                              "assistant")
        self.samdb.modify(msg)

        msg = ldb.Message()
        msg.dn = kept_dn
        msg["assistant"] = ldb.MessageElement("<GUID=%s>" % removed_guid,
                                              ldb.FLAG_MOD_ADD,
                                              "assistant")
        self.samdb.modify(msg)
        msg = ldb.Message()
        msg.dn = kept_dn
        msg["assistant"] = ldb.MessageElement("<GUID=%s>" % removed_guid,
                                              ldb.FLAG_MOD_DELETE,
                                              "assistant")
        self.samdb.modify(msg)

        #
        # Finally try the non-linked attribute 'assistant'
        # but with non existing GUID, SID, DN
        #
        msg = ldb.Message()
        msg.dn = kept_dn
        msg["assistant"] = ldb.MessageElement("CN=NoneNone,%s" % (basedn),
                                              ldb.FLAG_MOD_ADD,
                                              "assistant")
        try:
            self.samdb.modify(msg)
            self.fail("No exception should get LDB_ERR_CONSTRAINT_VIOLATION")
        except ldb.LdbError as e:
            (code, msg) = e.args
            self.assertEqual(code, ldb.ERR_CONSTRAINT_VIOLATION, str(e))
            werr = "%08X" % werror.WERR_DS_NAME_REFERENCE_INVALID
            self.assertTrue(werr in msg, msg)

        msg = ldb.Message()
        msg.dn = kept_dn
        msg["assistant"] = ldb.MessageElement("<SID=%s>" % none_sid_str,
                                              ldb.FLAG_MOD_ADD,
                                              "assistant")
        try:
            self.samdb.modify(msg)
            self.fail("No exception should get LDB_ERR_CONSTRAINT_VIOLATION")
        except ldb.LdbError as e:
            (code, msg) = e.args
            self.assertEqual(code, ldb.ERR_CONSTRAINT_VIOLATION, str(e))
            werr = "%08X" % werror.WERR_DS_NAME_REFERENCE_INVALID
            self.assertTrue(werr in msg, msg)

        msg = ldb.Message()
        msg.dn = kept_dn
        msg["assistant"] = ldb.MessageElement("<GUID=%s>" % none_guid_str,
                                              ldb.FLAG_MOD_ADD,
                                              "assistant")
        try:
            self.samdb.modify(msg)
            self.fail("No exception should get LDB_ERR_CONSTRAINT_VIOLATION")
        except ldb.LdbError as e:
            (code, msg) = e.args
            self.assertEqual(code, ldb.ERR_CONSTRAINT_VIOLATION, str(e))
            werr = "%08X" % werror.WERR_DS_NAME_REFERENCE_INVALID
            self.assertTrue(werr in msg, msg)

        self.samdb.delete(kept_dn)

    def test_normalize_dn_in_domain_full(self):
        domain_dn = self.samdb.domain_dn()

        part_dn = ldb.Dn(self.samdb, "CN=Users")

        full_dn = part_dn
        full_dn.add_base(domain_dn)

        full_str = str(full_dn)

        # That is, no change
        self.assertEqual(full_dn,
                         self.samdb.normalize_dn_in_domain(full_str))

    def test_normalize_dn_in_domain_part(self):
        domain_dn = self.samdb.domain_dn()

        part_str = "CN=Users"

        full_dn = ldb.Dn(self.samdb, part_str)
        full_dn.add_base(domain_dn)

        # That is, the domain DN appended
        self.assertEqual(full_dn,
                         self.samdb.normalize_dn_in_domain(part_str))

    def test_normalize_dn_in_domain_full_dn(self):
        domain_dn = self.samdb.domain_dn()

        part_dn = ldb.Dn(self.samdb, "CN=Users")

        full_dn = part_dn
        full_dn.add_base(domain_dn)

        # That is, no change
        self.assertEqual(full_dn,
                         self.samdb.normalize_dn_in_domain(full_dn))

    def test_normalize_dn_in_domain_part_dn(self):
        domain_dn = self.samdb.domain_dn()

        part_dn = ldb.Dn(self.samdb, "CN=Users")

        # That is, the domain DN appended
        self.assertEqual(ldb.Dn(self.samdb,
                                str(part_dn) + "," + str(domain_dn)),
                         self.samdb.normalize_dn_in_domain(part_dn))

class DsdbNCRootTests(TestCase):

    def setUp(self):
        super().setUp()
        self.lp = samba.tests.env_loadparm()
        self.creds = Credentials()
        self.creds.guess(self.lp)
        self.session = system_session()
        self.samdb = SamDB(session_info=self.session,
                           credentials=self.creds,
                           lp=self.lp)
        self.remote = False

    # These all use the local mode of operation inside
    # dsdb_find_nc_root() using the partitions control
    def test_dsdb_dn_nc_root_sid(self):
        dom_sid = self.samdb.get_domain_sid()
        domain_dn = ldb.Dn(self.samdb, self.samdb.domain_dn())
        dn = ldb.Dn(self.samdb, f"<SID={dom_sid}>")
        try:
            nc_root = self.samdb.get_nc_root(dn)
        except ldb.LdbError as e:
            (code, msg) = e.args
            self.fail("Got unexpected exception %d - %s "
                      % (code, msg))
        self.assertEqual(domain_dn, nc_root)

    def test_dsdb_dn_nc_root_admin_sid(self):
        dom_sid = self.samdb.get_domain_sid()
        domain_dn = ldb.Dn(self.samdb, self.samdb.domain_dn())
        dn = ldb.Dn(self.samdb, f"<SID={dom_sid}-500>")
        try:
            nc_root = self.samdb.get_nc_root(dn)
        except ldb.LdbError as e:
            (code, msg) = e.args
            self.fail("Got unexpected exception %d - %s "
                      % (code, msg))
        self.assertEqual(domain_dn, nc_root)

    def test_dsdb_dn_nc_root_users_container(self):
        dom_sid = self.samdb.get_domain_sid()
        domain_dn = ldb.Dn(self.samdb, self.samdb.domain_dn())
        dn = ldb.Dn(self.samdb, f"CN=Users,{domain_dn}")
        try:
            nc_root = self.samdb.get_nc_root(dn)
        except ldb.LdbError as e:
            (code, msg) = e.args
            self.fail("Got unexpected exception %d - %s "
                      % (code, msg))
        self.assertEqual(domain_dn, nc_root)

    def test_dsdb_dn_nc_root_new_dn(self):
        dom_sid = self.samdb.get_domain_sid()
        domain_dn = ldb.Dn(self.samdb, self.samdb.domain_dn())
        dn = ldb.Dn(self.samdb, f"CN=Xnotexisting,CN=Users,{domain_dn}")
        try:
            nc_root = self.samdb.get_nc_root(dn)
        except ldb.LdbError as e:
            (code, msg) = e.args
            self.fail("Got unexpected exception %d - %s "
                      % (code, msg))
        self.assertEqual(domain_dn, nc_root)

    def test_dsdb_dn_nc_root_new_dn_with_guid(self):
        domain_dn = ldb.Dn(self.samdb, self.samdb.domain_dn())
        dn = ldb.Dn(self.samdb, f"<GUID=828e3baf-fa02-4d82-ba5d-6f647dab5fd8>;CN=Xnotexisting,CN=Users,{domain_dn}")
        try:
            nc_root = self.samdb.get_nc_root(dn)
        except ldb.LdbError as e:
            (code, msg) = e.args
            self.fail("Got unexpected exception %d - %s "
                      % (code, msg))
        self.assertEqual(domain_dn, nc_root)

    def test_dsdb_dn_nc_root_guid(self):
        ntds_guid = self.samdb.get_ntds_GUID()
        configuration_dn = self.samdb.get_config_basedn()
        dn = ldb.Dn(self.samdb, f"<GUID={ntds_guid}>")
        try:
            nc_root = self.samdb.get_nc_root(dn)
        except ldb.LdbError as e:
            (code, msg) = e.args
            self.fail("Got unexpected exception %d - %s "
                      % (code, msg))
        self.assertEqual(configuration_dn, nc_root)

    def test_dsdb_dn_nc_root_misleading_to_noexisting_guid(self):
        ntds_guid = self.samdb.get_ntds_GUID()
        configuration_dn = self.samdb.get_config_basedn()
        domain_dn = ldb.Dn(self.samdb, self.samdb.domain_dn())
        dn = ldb.Dn(self.samdb, f"<GUID={ntds_guid}>;CN=Xnotexisting,CN=Users,{domain_dn}")
        try:
            nc_root = self.samdb.get_nc_root(dn)
        except ldb.LdbError as e:
            (code, msg) = e.args
            self.fail("Got unexpected exception %d - %s "
                      % (code, msg))
        self.assertEqual(configuration_dn, nc_root)

    def test_dsdb_dn_nc_root_misleading_to_existing_guid(self):
        ntds_guid = self.samdb.get_ntds_GUID()
        configuration_dn = self.samdb.get_config_basedn()
        domain_dn = ldb.Dn(self.samdb, self.samdb.domain_dn())
        dn = ldb.Dn(self.samdb, f"<GUID={ntds_guid}>;{domain_dn}")
        try:
            nc_root = self.samdb.get_nc_root(dn)
        except ldb.LdbError as e:
            (code, msg) = e.args
            self.fail("Got unexpected exception %d - %s "
                      % (code, msg))
        self.assertEqual(configuration_dn, nc_root)

class DsdbRemoteNCRootTests(DsdbNCRootTests):
    def setUp(self):
        super().setUp()
        # Reconnect to the remote LDAP port
        self.samdb = SamDB(url="ldap://%s" % samba.tests.env_get_var_value('SERVER'),
                           session_info=self.session,
                           credentials=self.get_credentials(),
                           lp=self.lp)
        self.remote = True


class DsdbFullScanTests(TestCase):

    def setUp(self):
        super().setUp()
        self.lp = samba.tests.env_loadparm()
        self.creds = Credentials()
        self.creds.guess(self.lp)
        self.session = system_session()

    def test_sam_ldb_open_no_full_scan(self):
        try:
            self.samdb = SamDB(session_info=self.session,
                               credentials=self.creds,
                               lp=self.lp,
                               options=["disable_full_db_scan_for_self_test:1"])
        except ldb.LdbError as err:
            estr = err.args[1]
            self.fail("sam.ldb required a full scan to start up")

class DsdbStartUpTests(TestCase):
    def setUp(self):
        super().setUp()
        lp = samba.tests.env_loadparm()
        path = lp.configfile

        # This is to avoid a tattoo of the global state
        self.lp = LoadParm(filename_for_non_global_lp=path)
        self.creds = Credentials()
        self.creds.guess(self.lp)
        self.session = system_session()
        self.samdb = SamDB(session_info=self.session,
                           credentials=self.creds,
                           lp=self.lp)

    def test_correct_fl(self):
        res = self.samdb.search(base="",
                                 scope=ldb.SCOPE_BASE,
                                 attrs=["domainFunctionality"])
        # This confirms the domain is in FL 2016 by default, this is
        # important to verify the original state
        self.assertEqual(int(res[0]["domainFunctionality"][0]),
                         dsdb.DS_DOMAIN_FUNCTION_2016)
        self.assertEqual(functional_level.dc_level_from_lp(self.lp),
                         dsdb.DS_DOMAIN_FUNCTION_2016)
        dsdb.check_and_update_fl(self.samdb, self.lp)

    def test_lower_smb_conf_fl(self):
        old_lp_fl = self.lp.get("ad dc functional level")
        self.lp.set("ad dc functional level",
                    "2008_R2")
        self.addCleanup(self.lp.set, "ad dc functional level", old_lp_fl)
        try:
            dsdb.check_and_update_fl(self.samdb, self.lp)
            self.fail("Should have failed to start DC with 2008 R2 FL in 2016 domain")
        except ldb.LdbError as err:
            (errno, estr) = err.args
            self.assertEqual(errno, ldb.ERR_CONSTRAINT_VIOLATION)



# Unix SMB/CIFS implementation.
#
# Claim type model.
#
# Copyright (C) Catalyst.Net Ltd. 2023
#
# Written by Rob van der Linde <rob@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import binascii
import os

from .fields import BooleanField, DnField, IntegerField,\
    PossibleClaimValuesField, StringField
from .model import Model
from .value_type import ValueType


class ClaimType(Model):
    enabled = BooleanField("Enabled")
    description = StringField("description")
    display_name = StringField("displayName")
    claim_attribute_source = DnField("msDS-ClaimAttributeSource")
    claim_is_single_valued = BooleanField("msDS-ClaimIsSingleValued")
    claim_is_value_space_restricted = BooleanField(
        "msDS-ClaimIsValueSpaceRestricted")
    claim_possible_values = PossibleClaimValuesField("msDS-ClaimPossibleValues")
    claim_source_type = StringField("msDS-ClaimSourceType")
    claim_type_applies_to_class = DnField(
        "msDS-ClaimTypeAppliesToClass", many=True)
    claim_value_type = IntegerField("msDS-ClaimValueType")

    def __str__(self):
        return str(self.display_name)

    @staticmethod
    def get_base_dn(samdb):
        """Return the base DN for the ClaimType model.

        :param samdb: SamDB connection
        :return: Dn object of container
        """
        base_dn = samdb.get_config_basedn()
        base_dn.add_child("CN=Claim Types,CN=Claims Configuration,CN=Services")
        return base_dn

    @staticmethod
    def get_object_class():
        return "msDS-ClaimType"

    @staticmethod
    def new_claim_type(samdb, attribute, applies_to, display_name=None,
                       description=None, enabled=True):
        """Creates a ClaimType but does not save the instance.

        :param samdb: SamDB database connection
        :param attribute: AttributeSchema object to use for creating ClaimType
        :param applies_to: List of ClassSchema objects ClaimType applies to
        :param display_name: Optional display name to use or use attribute name
        :param description: Optional description or fall back to display_name
        :param enabled: Create an enabled or disabled claim type (default True)
        :raises NotFound: if the ValueType for this attribute doesn't exist
        """
        value_type = ValueType.find(samdb, attribute)

        # Generate the new Claim Type cn.
        # Windows creates a random number here containing 16 hex digits.
        # We can achieve something similar using urandom(8)
        instance = binascii.hexlify(os.urandom(8)).decode()
        cn = f"ad://ext/{display_name}:{instance}"

        # if displayName is missing use attribute name.
        if display_name is None:
            display_name = attribute.name

        # adminDescription should be present but still have a fallback.
        if description is None:
            description = attribute.admin_description or display_name

        # claim_is_value_space_restricted is always False because we don't
        # yet support creating claims with a restricted possible values list.
        return ClaimType(
            cn=cn,
            description=description,
            display_name=display_name,
            enabled=enabled,
            claim_attribute_source=attribute.dn,
            claim_is_single_valued=attribute.is_single_valued,
            claim_is_value_space_restricted=False,
            claim_source_type="AD",
            claim_type_applies_to_class=[obj.dn for obj in applies_to],
            claim_value_type=value_type.claim_value_type,
        )



# Unix SMB/CIFS implementation.
# Copyright (C) Andrew Bartlett <abartlet@samba.org> 2016
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for array handling in PIDL generated bindings samba.dcerpc.*"""

from samba.dcerpc import drsblobs
import samba.tests
from samba.ndr import ndr_unpack, ndr_pack
import talloc
import gc


class ArrayTests(samba.tests.TestCase):

    def setUp(self):
        super().setUp()
        talloc.enable_null_tracking()
        self.startup_blocks = talloc.total_blocks()

    def tearDown(self):
        super().tearDown()
        gc.collect()
        if talloc.total_blocks() != self.startup_blocks:
            talloc.report_full()
            self.fail("it appears we are leaking memory")

    def test_array_from_python(self):
        rmd = drsblobs.replPropertyMetaDataBlob()

        rmd.version = 1
        rmd.ctr = drsblobs.replPropertyMetaDataCtr1()
        rmd.ctr.count = 3

        rmd1 = drsblobs.replPropertyMetaData1()
        rmd1.attid = 1
        rmd1.version = 2

        rmd2 = drsblobs.replPropertyMetaData1()
        rmd2.attid = 2
        rmd2.version = 2

        rmd3 = drsblobs.replPropertyMetaData1()
        rmd3.attid = 3
        rmd3.version = 2

        rmd.ctr.array = [rmd1, rmd2, rmd3]
        gc.collect()

        self.assertIsNotNone(rmd)
        self.assertEqual(rmd.version, 1)
        self.assertIsNotNone(rmd.ctr)
        self.assertEqual(rmd.ctr.count, 3)
        self.assertEqual(len(rmd.ctr.array), rmd.ctr.count)
        self.assertIsNotNone(rmd.ctr.array[0])
        self.assertEqual(rmd.ctr.array[0].attid, 1)

    def test_array_with_exception(self):
        try:
            rmd = drsblobs.replPropertyMetaDataBlob()

            rmd.version = 1
            rmd.ctr = drsblobs.replPropertyMetaDataCtr1()
            rmd.ctr.count = 3

            rmd1 = drsblobs.replPropertyMetaData1()
            rmd1.attid = 1
            rmd1.version = 2

            rmd2 = drsblobs.replPropertyMetaData1()
            rmd2.attid = 2
            rmd2.version = 2

            rmd3 = drsblobs.replPropertyMetaData1()
            rmd3.attid = 3
            rmd3.version = 2

            rmd.ctr.array = [rmd1, rmd2, rmd3]

            gc.collect()

            self.assertIsNotNone(rmd)
            self.assertEqual(rmd.version, 1)
            self.assertIsNotNone(rmd.ctr)
            self.assertEqual(rmd.ctr.count, 3)
            self.assertEqual(len(rmd.ctr.array), rmd.ctr.count)
            self.assertIsNotNone(rmd.ctr.array[0])
            self.assertEqual(rmd.ctr.array[0].attid, 1)

            raise Exception()
        except:
            pass

    def test_array_from_python_function(self):
        def get_rmd():
            rmd = drsblobs.replPropertyMetaDataBlob()

            rmd.version = 1
            rmd.ctr = drsblobs.replPropertyMetaDataCtr1()
            rmd.ctr.count = 3

            rmd1 = drsblobs.replPropertyMetaData1()
            rmd1.attid = 1
            rmd1.version = 2

            rmd2 = drsblobs.replPropertyMetaData1()
            rmd2.attid = 2
            rmd2.version = 2

            rmd3 = drsblobs.replPropertyMetaData1()
            rmd3.attid = 3
            rmd3.version = 2

            rmd.ctr.array = [rmd1, rmd2, rmd3]
            return rmd

        rmd = get_rmd()
        gc.collect()
        self.assertIsNotNone(rmd)
        self.assertEqual(rmd.version, 1)
        self.assertIsNotNone(rmd.ctr)
        self.assertEqual(rmd.ctr.count, 3)
        self.assertEqual(len(rmd.ctr.array), rmd.ctr.count)
        self.assertIsNotNone(rmd.ctr.array[0])
        self.assertEqual(rmd.ctr.array[0].attid, 1)

    def test_array_from_ndr(self):
        rmd = drsblobs.replPropertyMetaDataBlob()

        rmd.version = 1
        rmd.ctr = drsblobs.replPropertyMetaDataCtr1()
        rmd.ctr.count = 3

        rmd1 = drsblobs.replPropertyMetaData1()
        rmd1.attid = 1
        rmd1.version = 2

        rmd2 = drsblobs.replPropertyMetaData1()
        rmd2.attid = 2
        rmd2.version = 2

        rmd3 = drsblobs.replPropertyMetaData1()
        rmd3.attid = 3
        rmd3.version = 2

        rmd.ctr.array = [rmd1, rmd2, rmd3]

        packed = ndr_pack(rmd)
        gc.collect()

        rmd_unpacked = ndr_unpack(drsblobs.replPropertyMetaDataBlob, packed)
        self.assertIsNotNone(rmd_unpacked)
        self.assertEqual(rmd_unpacked.version, 1)
        self.assertIsNotNone(rmd_unpacked.ctr)
        self.assertEqual(rmd_unpacked.ctr.count, 3)
        self.assertEqual(len(rmd_unpacked.ctr.array), rmd_unpacked.ctr.count)
        self.assertIsNotNone(rmd_unpacked.ctr.array[0])
        self.assertEqual(rmd_unpacked.ctr.array[0].attid, 1)

        self.assertEqual(rmd.ctr.array[0].attid,
                         rmd_unpacked.ctr.array[0].attid)

    def test_array_delete(self):
        rmd = drsblobs.replPropertyMetaDataBlob()

        rmd.version = 1
        rmd.ctr = drsblobs.replPropertyMetaDataCtr1()
        rmd.ctr.count = 3

        rmd1 = drsblobs.replPropertyMetaData1()
        rmd1.attid = 1
        rmd1.version = 2

        rmd2 = drsblobs.replPropertyMetaData1()
        rmd2.attid = 2
        rmd2.version = 2

        rmd3 = drsblobs.replPropertyMetaData1()
        rmd3.attid = 3
        rmd3.version = 2

        rmd.ctr.array = [rmd1, rmd2, rmd3]
        try:
            del rmd1.version
            self.fail("succeeded in deleting rmd1.version")
        except AttributeError as e:
            pass

        try:
            del rmd.ctr.array
            self.fail("succeeded in deleting rmd.ctr.array")
        except AttributeError as e:
            pass



# Unix SMB/CIFS implementation.
#
# Copyright (C) Andrew Bartlett 2011
# Copyright (C) Isaac Boukris 2020
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for the CreateTrustedDomainRelax wrapper"""

import samba
from samba.tests import TestCase
from samba.dcerpc import lsa, security, drsblobs
from samba.credentials import (
    Credentials,
    SMB_ENCRYPTION_REQUIRED,
    SMB_ENCRYPTION_OFF
)
from samba.lsa_utils import (
    OpenPolicyFallback,
    CreateTrustedDomainRelax,
    CreateTrustedDomainFallback,
)


class CreateTrustedDomain(TestCase):
    smbencrypt = True

    def get_user_creds(self):
        c = Credentials()
        c.guess()
        domain = samba.tests.env_get_var_value('DOMAIN')
        username = samba.tests.env_get_var_value('USERNAME')
        password = samba.tests.env_get_var_value('PASSWORD')
        c.set_domain(domain)
        c.set_username(username)
        c.set_password(password)
        return c

    def new_lsa_conn(self, basis_connection=None):
        creds = self.get_user_creds()
        if self.smbencrypt:
            creds.set_smb_encryption(SMB_ENCRYPTION_REQUIRED)
        else:
            creds.set_smb_encryption(SMB_ENCRYPTION_OFF)

        lp = self.get_loadparm()
        binding_string = (
            "ncacn_np:%s" % (samba.tests.env_get_var_value('SERVER'))
        )

        lsa_conn = lsa.lsarpc(
            binding_string,
            lp,
            creds,
            basis_connection=basis_connection
        )

        if self.smbencrypt:
            self.assertTrue(lsa_conn.transport_encrypted())
        else:
            self.assertFalse(lsa_conn.transport_encrypted())

        return lsa_conn

    def _create_trust_relax(self, smbencrypt=True):
        self.smbencrypt = smbencrypt

        in_version = 1
        in_revision_info1 = lsa.revision_info1()
        in_revision_info1.revision = 1
        in_revision_info1.supported_features = (
            lsa.LSA_FEATURE_TDO_AUTH_INFO_AES_CIPHER
        )

        (
            lsa_conn,
            out_version,
            out_revision_info1,
            pol_handle
        ) = OpenPolicyFallback(
            self.new_lsa_conn,
            '',
            in_version,
            in_revision_info1,
            False,
            security.SEC_FLAG_MAXIMUM_ALLOWED
        )
        self.assertIsNotNone(pol_handle)

        name = lsa.String()
        name.string = "tests.samba.example.com"
        try:
            info = lsa_conn.QueryTrustedDomainInfoByName(
                pol_handle,
                name,
                lsa.LSA_TRUSTED_DOMAIN_INFO_FULL_INFO
            )

            lsa_conn.DeleteTrustedDomain(pol_handle, info.info_ex.sid)
        except RuntimeError:
            pass

        info = lsa.TrustDomainInfoInfoEx()
        info.domain_name.string = name.string
        info.netbios_name.string = "TESTTRUSTRELAXX"
        info.sid = security.dom_sid("S-1-5-21-538490383-3740119673-95748416")
        info.trust_direction = (
            lsa.LSA_TRUST_DIRECTION_INBOUND
            | lsa.LSA_TRUST_DIRECTION_OUTBOUND
        )
        info.trust_type = lsa.LSA_TRUST_TYPE_UPLEVEL
        info.trust_attributes = lsa.LSA_TRUST_ATTRIBUTE_FOREST_TRANSITIVE

        password_blob = list("password".encode('utf-16-le'))

        clear_value = drsblobs.AuthInfoClear()
        clear_value.size = len(password_blob)
        clear_value.password = password_blob

        clear_authentication_information = drsblobs.AuthenticationInformation()
        clear_authentication_information.LastUpdateTime = 0
        clear_authentication_information.AuthType = lsa.TRUST_AUTH_TYPE_CLEAR
        clear_authentication_information.AuthInfo = clear_value

        auth_info_array = drsblobs.AuthenticationInformationArray()
        auth_info_array.count = 1
        auth_info_array.array = [clear_authentication_information]

        outgoing = drsblobs.trustAuthInOutBlob()
        outgoing.count = 1
        outgoing.current = auth_info_array

        trustdom_handle = None
        try:
            trustdom_handle = CreateTrustedDomainRelax(lsa_conn,
                                                       pol_handle,
                                                       info,
                                                       security.SEC_STD_DELETE,
                                                       outgoing,
                                                       outgoing)
        except samba.NTSTATUSError as nt:
            raise AssertionError(nt)
        except OSError as e:
            if smbencrypt:
                raise AssertionError(e)

        if smbencrypt:
            self.assertIsNotNone(trustdom_handle)
            lsa_conn.DeleteTrustedDomain(pol_handle, info.sid)
        else:
            self.assertIsNone(trustdom_handle)

    def _create_trust_fallback(self):
        self.smbencrypt = True

        in_version = 1
        in_revision_info1 = lsa.revision_info1()
        in_revision_info1.revision = 1
        in_revision_info1.supported_features = (
            lsa.LSA_FEATURE_TDO_AUTH_INFO_AES_CIPHER
        )

        (
            lsa_conn,
            out_version,
            out_revision_info1,
            pol_handle
        ) = OpenPolicyFallback(
            self.new_lsa_conn,
            '',
            in_version,
            in_revision_info1,
            False,
            security.SEC_FLAG_MAXIMUM_ALLOWED
        )
        self.assertIsNotNone(pol_handle)

        name = lsa.String()
        name.string = "tests.samba.example.com"
        try:
            info = lsa_conn.QueryTrustedDomainInfoByName(
                pol_handle,
                name,
                lsa.LSA_TRUSTED_DOMAIN_INFO_FULL_INFO
            )

            lsa_conn.DeleteTrustedDomain(pol_handle, info.info_ex.sid)
        except RuntimeError:
            pass

        info = lsa.TrustDomainInfoInfoEx()
        info.domain_name.string = name.string
        info.netbios_name.string = "TESTTRUSTRELAXX"
        info.sid = security.dom_sid("S-1-5-21-538490383-3740119673-95748416")
        info.trust_direction = (
            lsa.LSA_TRUST_DIRECTION_INBOUND
            | lsa.LSA_TRUST_DIRECTION_OUTBOUND
        )
        info.trust_type = lsa.LSA_TRUST_TYPE_UPLEVEL
        info.trust_attributes = lsa.LSA_TRUST_ATTRIBUTE_FOREST_TRANSITIVE

        password_blob = list("password".encode('utf-16-le'))

        clear_value = drsblobs.AuthInfoClear()
        clear_value.size = len(password_blob)
        clear_value.password = password_blob

        clear_authentication_information = drsblobs.AuthenticationInformation()
        clear_authentication_information.LastUpdateTime = 0
        clear_authentication_information.AuthType = lsa.TRUST_AUTH_TYPE_CLEAR
        clear_authentication_information.AuthInfo = clear_value

        auth_info_array = drsblobs.AuthenticationInformationArray()
        auth_info_array.count = 1
        auth_info_array.array = [clear_authentication_information]

        outgoing = drsblobs.trustAuthInOutBlob()
        outgoing.count = 1
        outgoing.current = auth_info_array

        trustdom_handle = None
        try:
            trustdom_handle = CreateTrustedDomainFallback(
                lsa_conn,
                pol_handle,
                info,
                security.SEC_STD_DELETE,
                out_version,
                out_revision_info1,
                outgoing,
                outgoing
            )
        except samba.NTSTATUSError as nt:
            raise AssertionError(nt)

        self.assertIsNotNone(trustdom_handle)
        lsa_conn.DeleteTrustedDomain(pol_handle, info.sid)

    def test_create_trust_relax_encrypt(self):
        self._create_trust_relax(True)

    def test_create_trust_relax_no_enc(self):
        self._create_trust_relax(False)

    def test_create_trust_fallback(self):
        self._create_trust_fallback()



# Unix SMB/CIFS implementation.
#
# Samba domain models.
#
# Copyright (C) Catalyst.Net Ltd. 2023
#
# Written by Rob van der Linde <rob@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .auth_policy import (AuthenticationPolicy, StrongNTLMPolicy,
                          MIN_TGT_LIFETIME, MAX_TGT_LIFETIME)
from .auth_silo import AuthenticationSilo
from .claim_type import ClaimType
from .computer import Computer
from .container import Container
from .gmsa import GroupManagedServiceAccount
from .group import Group
from .model import Model
from .org import OrganizationalPerson, OrganizationalUnit
from .person import Person
from .registry import MODELS
from .schema import AttributeSchema, ClassSchema
from .site import Site
from .subnet import Subnet
from .types import (AccountType, GroupType, SupportedEncryptionTypes,
                    SystemFlags, UserAccountControl)
from .user import User
from .value_type import ValueType



# Blackbox tests for http_test
#
# Copyright (C) Noel Power noel.power@suse.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import time
import threading
import logging
import json
from http.server import HTTPServer, BaseHTTPRequestHandler
from samba.logger import get_samba_logger
from samba.tests import BlackboxTestCase, BlackboxProcessError

logger = get_samba_logger(name=__name__)
COMMAND = "bin/http_test"
def make_chunks(msg, chunk_size):
    chunks = []
    while len(msg) > chunk_size:
        chunk = msg[:chunk_size]
        chunks.append(chunk)
        msg = msg[chunk_size:]
    if len(msg):
        chunks.append(msg)
    return chunks

# simple handler, spits back the 'path' passed in
# GET or POST and a chunked encoded http response
# where the chunk size is 10 octets
class ChunkHTTPRequestHandler(BaseHTTPRequestHandler):
    def handle_req(self):
        msg = bytes(self.path, encoding="utf-8")
        chunks = make_chunks(msg, 10)

        self.send_response(200)
        self.send_header('content-type', 'application/json; charset=UTF-8')
        if self.path == "usegziptransferencoding":
            self.send_header('Transfer-Encoding', 'gzip')
        else:
            self.send_header('Transfer-Encoding', 'chunked')
        self.end_headers()
        resp = bytes()
        for chunk in chunks:
            resp = resp + ("%x" % len(chunk)).encode("utf-8") + b'\r\n' + chunk + b'\r\n'
        resp += b'0\r\n\r\n'
        self.wfile.write(resp)

    def do_POST(self):
        self.handle_req()
    def do_GET(self):
        self.handle_req()

class HttpChunkBlackboxTests(BlackboxTestCase):
    def setUp(self):
        self.server = HTTPServer((os.getenv("SERVER_IP", "localhost"), 8080),
                                 ChunkHTTPRequestHandler,
                                 bind_and_activate=False)
        self.t = threading.Thread(target=HttpChunkBlackboxTests.http_server, args=(self,))
        self.t.setDaemon(True)
        self.t.start()
        time.sleep(1)

    def tearDown(self):
        super().tearDown()

    def http_server(self):
        self.server.server_bind()
        self.server.server_activate()
        self.server.serve_forever()

    def test_single_chunk(self):
        try:
            msg = "one_chunk"
            resp = self.check_output("%s -U%% -I%s --uri %s" % (COMMAND, os.getenv("SERVER_IP", "localhost"), msg))
            self.assertEqual(msg,resp.decode('utf-8'))
        except BlackboxProcessError as e:
            print("Failed with: %s" % e)
            self.fail(str(e))

    def test_multi_chunks(self):
        try:
            msg = "snglechunksnglechunksnglechunksnglechunksnglechunk"
            resp = self.check_output("%s -U%% -I%s --uri %s" % (COMMAND, os.getenv("SERVER_IP", "localhost"), msg))
            self.assertEqual(msg, resp.decode('utf-8'))
        except BlackboxProcessError as e:
            print("Failed with: %s" % e)
            self.fail(str(e))

    def test_exceed_request_size(self):
        try:
            msg = "snglechunksnglechunksnglechunksnglechunksnglechunk"
            resp = self.check_output("%s -d11 -U%% -I%s --rsize 49 --uri %s" % (COMMAND, os.getenv("SERVER_IP", "localhost"), msg))
            self.fail("unexpected success")
        except BlackboxProcessError as e:
            if "http_read_chunk: size 50 exceeds max content len 49 skipping body" not in e.stderr.decode('utf-8'):
                self.fail(str(e))
            if "unexpected 0 len response" not in e.stdout.decode('utf-8'):
                self.fail(str(e))

    def test_exact_request_size(self):
        try:
            msg = "snglechunksnglechunksnglechunksnglechunksnglechunk"
            resp = self.check_output("%s -U%% -I%s --rsize 50 --uri %s" % (COMMAND, os.getenv("SERVER_IP", "localhost"), msg))
            self.assertEqual(msg, resp.decode('utf-8'))
        except BlackboxProcessError as e:
            print("Failed with: %s" % e)
            self.fail(str(e))

    def test_gzip_transfer_encoding(self):
        try:
            msg = "usegziptransferencoding"
            resp = self.check_output("%s -U%% -I%s --rsize 50 --uri %s" % (COMMAND, os.getenv("SERVER_IP", "localhost"), msg))
            self.assertEqual(msg, resp.decode('utf-8'))
            self.fail("unexpected success")
        except BlackboxProcessError as e:
            if "http_response_needs_body: Unsupported transfer encoding type gzip" not in e.stderr.decode('utf-8'):
                self.fail(str(e))



# Unix SMB/CIFS implementation.
# Copyright (C) Jelmer Vernooij <jelmer@samba.org> 2007
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for the Auth Python bindings.

Note that this just tests the bindings work. It does not intend to test
the functionality, that's already done in other tests.
"""

from samba import auth, ntstatus, NTSTATUSError
from samba.samdb import SamDB
import samba.tests


class AuthSystemSessionTests(samba.tests.TestCase):

    def setUp(self):
        super().setUp()
        self.system_session = auth.system_session()
        self.lp = samba.tests.env_loadparm()

    def test_system_session_attrs(self):
        self.assertTrue(hasattr(self.system_session, 'credentials'))
        self.assertTrue(hasattr(self.system_session, 'info'))
        self.assertTrue(hasattr(self.system_session, 'security_token'))
        self.assertTrue(hasattr(self.system_session, 'session_key'))
        self.assertTrue(hasattr(self.system_session, 'torture'))

    def test_system_session_credentials(self):
        self.assertIsNone(self.system_session.credentials.get_bind_dn())
        self.assertIsNotNone(self.system_session.credentials.get_password())
        self.assertEqual(self.system_session.credentials.get_username(),
                         self.lp.get('netbios name').upper() + "$")

    def test_system_session_info(self):
        self.assertEqual(self.system_session.info.full_name, 'System')
        self.assertEqual(self.system_session.info.domain_name, 'NT AUTHORITY')
        self.assertEqual(self.system_session.info.account_name, 'SYSTEM')

    def test_system_session_session_key(self):
        expected = b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        self.assertEqual(self.system_session.session_key, expected)

    def test_system_session_security_token(self):
        self.assertTrue(self.system_session.security_token.is_system())
        self.assertFalse(self.system_session.security_token.is_anonymous())


class AuthAdminSessionTests(samba.tests.TestCase):

    def setUp(self):
        super().setUp()
        self.lp = samba.tests.env_loadparm()
        self.admin_session = auth.admin_session(self.lp,
                                                "S-1-5-21-2212615479-2695158682-2101375467")

    def test_admin_session_attrs(self):
        self.assertTrue(hasattr(self.admin_session, 'credentials'))
        self.assertTrue(hasattr(self.admin_session, 'info'))
        self.assertTrue(hasattr(self.admin_session, 'security_token'))
        self.assertTrue(hasattr(self.admin_session, 'session_key'))
        self.assertTrue(hasattr(self.admin_session, 'torture'))

    def test_admin_session_credentials(self):
        self.assertIsNone(self.admin_session.credentials)

    def test_session_info_details(self):
        self.assertEqual(self.admin_session.info.full_name,
                         'Administrator')
        self.assertEqual(self.admin_session.info.domain_name,
                         self.lp.get('workgroup'))
        self.assertEqual(self.admin_session.info.account_name,
                         'Administrator')

    def test_security_token(self):
        self.assertFalse(self.admin_session.security_token.is_system())
        self.assertFalse(self.admin_session.security_token.is_anonymous())
        self.assertTrue(self.admin_session.security_token.has_builtin_administrators())

    def test_session_info_unix_details(self):
        samba.auth.session_info_fill_unix(session_info=self.admin_session,
                                          lp_ctx=self.lp,
                                          user_name="Administrator")
        self.assertEqual(self.admin_session.unix_info.sanitized_username,
                         'Administrator')
        self.assertEqual(self.admin_session.unix_info.unix_name,
                         self.lp.get('workgroup').upper() +
                         self.lp.get('winbind separator') + 'Administrator')
        self.assertIsNotNone(self.admin_session.unix_token)

    def test_user_session_principals(self):
        session = auth.system_session()
        realm = session.credentials.get_realm()
        samdb = SamDB(lp=self.lp, session_info=session)
        samdb.newuser('$$', 'password123!')
        self.addCleanup(samdb.deleteuser, '$$')
        for p, expected, upnc in [
                ('', ntstatus.NT_STATUS_NO_SUCH_USER, None),
                ('Administrator', ntstatus.NT_STATUS_NO_SUCH_USER, None),
                (f'Administrator@{realm.split(".", 2)[-1]}', ntstatus.NT_STATUS_NO_SUCH_USER, None),
                (f'Administrator\n@{realm}', ntstatus.NT_STATUS_NO_SUCH_USER, None),
                (f'Administrator@localdc.{realm}', ntstatus.NT_STATUS_NO_SUCH_USER, None),
                (f'Administrator@{realm.lower()}', 0, True),
                (f'administrator @ {realm}', 0, True),
                (f'JOE@{realm.lower()}', 0, False),
                (f'joe @{realm}', 0, False),
                (f'joe@ {realm.title()}', 0, False),
                (f'  joe      @  {realm} ', 0, False),
                (f'joe$@{realm}', ntstatus.NT_STATUS_NO_SUCH_USER, None),
                (f'@$@{realm} ', ntstatus.NT_STATUS_NO_SUCH_USER, None),
                (f'@{realm}', ntstatus.NT_STATUS_NO_SUCH_USER, None),
                (f'$$@{realm}', 0, False),
                (f'$@{realm}', ntstatus.NT_STATUS_NO_SUCH_USER, None),
                (f'localdc@{realm}', ntstatus.NT_STATUS_NO_SUCH_USER, None),
                (f'localdc$@{realm}', 0, True),
                (f'localdc.{realm}', ntstatus.NT_STATUS_NO_SUCH_USER, None),
                (f'{realm}', ntstatus.NT_STATUS_NO_SUCH_USER, None),
                (f'LOCALDC$@{realm}', 0, True),
                (f'missing$@{realm}', ntstatus.NT_STATUS_NO_SUCH_USER, None),
                ('localdc$', ntstatus.NT_STATUS_NO_SUCH_USER, None),
        ]:
            with self.subTest(p=p):
                try:
                    session = auth.user_session(samdb, lp_ctx=self.lp, principal=p)
                except NTSTATUSError as e:
                    result = e.args[0]
                else:
                    # no failure
                    self.assertEqual(session.info.user_principal_constructed, upnc, p)
                    result = 0

                self.assertEqual(result, expected, p)
